/*
linux/fs/dmsdos/dmsdos_dec.c

DMSDOS filesystem: decompression routines

******************************************************************************
DMSDOS (Doublespace/Drivespace compressed MSDOS filesystem) for Linux
written 1995,1996 by Frank Gockel

    (C) Copyright 1995,1996 by Frank Gockel

Some code of the dmsdos filesystem has been copied from the msdos filesystem
so there are the following additional copyrights:

    (C) Copyright 1992,1993 by Werner Almesberger (msdos filesystem)
    (C) Copyright 1994,1995 by Jacques Gelinas (mmap code)
    (C) Copyright 1992-1995 by Linus Torvalds

The DMSDOS filesystem was inspired by the THS filesystem (a simple doublespace
DS-0-2 compressed read-only filesystem) written 1994 by Thomas Scheuermann.

The DMSDOS filesystem is distributed under the Gnu General Public Licence.
See file COPYING for details.
******************************************************************************
                                                        
*/

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/stat.h>
#include <linux/mm.h>
#include <linux/locks.h>
#include <linux/fs.h>
#include <linux/malloc.h>
#include <linux/msdos_fs.h>
#include <linux/dmsdos_fs.h>

#include <asm/system.h>
#include <asm/segment.h>
#include <asm/bitops.h>

extern Dblsb dblsb[];

/* known compression methods */
/* see header file */

/* some functions are inline for speed reasons - this eats kernel memory */

inline int getbit(unsigned char*clusterk,int*mask,int*count)
{ int ret;

  ret=(*mask)&clusterk[*count];
  (*mask)<<=1;
  if(*mask==256)
  { *mask=1;
    (*count)++;
  }  
  return ret;
}

int getbits(unsigned char*clusterk,int*mask,int*count,int bits)
{ int orwert=1;
  int ret=0;
  int i;
  
  for(i=0;i<bits;++i)
  { if(getbit(clusterk,mask,count))ret|=orwert;
    orwert<<=1;
  }
  return ret;
}

int getDX(unsigned char*clusterk,int*mask,int*count)
{ if(!getbit(clusterk,mask,count)) return getbits(clusterk,mask,count,6);
  if(!getbit(clusterk,mask,count)) return getbits(clusterk,mask,count,8)+64;
  return getbits(clusterk,mask,count,12)+320;
}

int getCX(unsigned char*clusterk,int*mask,int*count)
{ if(getbit(clusterk,mask,count)) return 3;
  if(getbit(clusterk,mask,count)) return getbits(clusterk,mask,count,1)+4;
  if(getbit(clusterk,mask,count)) return getbits(clusterk,mask,count,2)+6;
  if(getbit(clusterk,mask,count)) return getbits(clusterk,mask,count,3)+10;
  if(getbit(clusterk,mask,count)) return getbits(clusterk,mask,count,4)+18;
  if(getbit(clusterk,mask,count)) return getbits(clusterk,mask,count,5)+34;
  if(getbit(clusterk,mask,count)) return getbits(clusterk,mask,count,6)+66;
  if(getbit(clusterk,mask,count)) return getbits(clusterk,mask,count,7)+130;
  if(getbit(clusterk,mask,count)) return getbits(clusterk,mask,count,8)+258;
  return -1;
}

int decrb(unsigned char*clusterk,int*mask,int*count,
          unsigned char*clusterd,int*pos,int*sek,int dx,int k)
{ int i,cx;

  if(dx==0)return -2;
  if(dx==0x113f)
  { LOG_DECOMP("DMSDOS: decrb: 0x113f sync found.\n");
    (*sek)--;
    if(*sek==0)return -1;
    if(*pos%SECTOR_SIZE) return -2;
    return 0;
  }
  cx=getCX(clusterk,mask,count)+k;
  if(cx<=0)return -2;
  if(dx>*pos) return -2;
  for(i=0;i<cx;i++)clusterd[(*pos)+i]=clusterd[(*pos)+i-dx];
  (*pos)+=cx;
  return 0;
}

/* decompress a compressed doublespace/drivespace cluster clusterk to clusterd
*/
int dbl_decompress(unsigned char*clusterd, unsigned char*clusterk,
               Mdfat_entry*mde)
{
	int sekcount;
	int r,count,mask,pos,dx;

	sekcount=mde->size_hi_minus_1+1;
	count=0;
	mask=1;
	pos=0;

  switch(getbits(clusterk,&mask,&count,32))
  {
    case DS_0_0:
    case DS_0_1:
    case DS_0_2:
    LOG_DECOMP("DMSDOS: decompressing DS-0-x\n");
        do
        { r=0;
          if(!getbit(clusterk,&mask,&count))
          { if(getbit(clusterk,&mask,&count))
            { clusterd[pos++]=getbits(clusterk,&mask,&count,7);
            }
            else
            { dx=getbits(clusterk,&mask,&count,6);
              r=decrb(clusterk,&mask,&count,clusterd,&pos,&sekcount,dx,-1);
            }
          }
          else
          { if(!getbit(clusterk,&mask,&count))
            { clusterd[pos++]=getbits(clusterk,&mask,&count,7)|128;
            }
            else
            { dx=(!getbit(clusterk,&mask,&count)) ?
                     getbits(clusterk,&mask,&count,8)+64 :
                     getbits(clusterk,&mask,&count,12)+320;
              r=decrb(clusterk,&mask,&count,clusterd,&pos,&sekcount,dx,-1);
            }
          }
        }
        while(r==0);
        if(r==-2)
        { printk("DMSDOS: error in DS-0-x compressed data.\n");
          return -2;
        }	
        LOG_DECOMP("DMSDOS: decompress finished.\n");
	return 0;
	
    case JM_0_0:
    case JM_0_1:
    LOG_DECOMP("DMSDOS: decompressing JM-0-x\n");
	do
	{ r=0;
	  if(!getbit(clusterk,&mask,&count))
	  { clusterd[pos++]=getbits(clusterk,&mask,&count,7);
	  }
	  else
	  { if(!getbit(clusterk,&mask,&count))
	    { dx=getDX(clusterk,&mask,&count);
	      r=decrb(clusterk,&mask,&count,clusterd,&pos,&sekcount,dx,0);
	    }
	    else
	    { clusterd[pos++]=getbits(clusterk,&mask,&count,7)|128;
	    }
	  }
	}
	while(r==0);
	if(r==-2)
	{ printk("DMSDOS: error in JM-0-x compressed data.\n");
	  return -2;
	}
	LOG_DECOMP("DMSDOS: decompress finished.\n");
	return 0;

    case SQ_0_0:
        printk("DMSDOS: Sorry, SQ-0-0 compression not yet supported.\n");    
        return -1;
        
    default:	
	printk("DMSDOS: compression method not recognized.\n");
	return -1;
	
  } /* end switch */
  
  return 0;
}

/* read a complete file cluster and decompress it if necessary; 
   this function is unable to read cluster 0 (CVF root directory) */
/* returns cluster length in bytes or error (<0) */
/* this function is specific to doublespace/drivespace */
int dbl_read_cluster(struct super_block*sb,
                        unsigned char*clusterd, int clusternr,int cvfnr)
{	Mdfat_entry mde;
	unsigned char*clusterk;
	int anz_sektoren;
	int i;
	struct buffer_head*bh;
	int membytes;
	int sektor;
	/*int fullsize;*/

	LOG_CLUST("DMSDOS: dbl_read_cluster %d in CVF %d\n",clusternr,cvfnr+1);

	dbl_mdfat_value(sb,clusternr,cvfnr,NULL,&mde);
	
	if((mde.flags&2)==0)
	{ /* hmm, cluster is unused (it's a lost or ghost cluster)
	     and contains undefined data, but it *is* readable */
	  /* oh no, it contains ZEROD data per definition...
	     this is really important */
	  memset(clusterd,0,dblsb[cvfnr].s_sectperclust*SECTOR_SIZE);
	  LOG_CLUST("DMSDOS: lost cluster %d in CVF %d detected\n",clusternr,
	         cvfnr+1);
	  return 0; /* yes, has length zero */
	}
	
        sektor=mde.sector_minus_1+1;
        anz_sektoren=mde.size_lo_minus_1+1;/* real sectors on disk */
        if(anz_sektoren>dblsb[cvfnr].s_sectperclust)
        { printk("DMSDOS: read_cluster: mdfat sektors > sectperclust, cutting\n");
          anz_sektoren=dblsb[cvfnr].s_sectperclust;
        }
        
        if(mde.flags&1)
        {  /* cluster is not compressed */
	   for(i=0;i<anz_sektoren;++i)
	   {  bh=read_dbl_sector(sb,sektor+i,cvfnr);
	      if(bh==NULL){UNLOCK_READWRITE;return -EIO;}
	      memcpy(&clusterd[i*SECTOR_SIZE],bh->b_data,SECTOR_SIZE);
	      bh_free(sb,bh);
	   }
        }
        else
        {  /* cluster is compressed */
           
           membytes=SECTOR_SIZE*anz_sektoren;
           
           clusterk=(unsigned char*)MALLOC(membytes);
           if(clusterk==NULL)
           { printk("DMSDOS: no memory for decompression!\n");
             return -2;
           }
           
           for(i=0;i<anz_sektoren;++i)
	   {  bh=read_dbl_sector(sb,sektor+i,cvfnr);
	      if(bh==NULL)
	      { FREE(clusterk);
	        return -EIO;
	      }
	      memcpy(&clusterk[i*SECTOR_SIZE],bh->b_data,SECTOR_SIZE);
	      bh_free(sb,bh);
	   }
	   
           i=dbl_decompress(clusterd,clusterk,&mde);
           
           FREE(clusterk);
           
           if(i)
           { printk("DMSDOS: decompression of cluster %d in CVF %d failed.\n",
                    clusternr,cvfnr+1);
             return i;
           }
           
        }

        /* the slack must be zerod out */
        if(mde.size_hi_minus_1+1<dblsb[cvfnr].s_sectperclust)
        { memset(clusterd+(mde.size_hi_minus_1+1)*SECTOR_SIZE,0,
                  (dblsb[cvfnr].s_sectperclust-mde.size_hi_minus_1-1)*
                                                              SECTOR_SIZE);
        }
        
	return (mde.size_hi_minus_1+1)*SECTOR_SIZE;
}

/* read a complete file cluster and decompress it if necessary; 
   it must be able to read directories
   this function is unable to read cluster 0 (CVF root directory) */
/* returns cluster length in bytes or error (<0) */
/* this function is a generic wrapper */
int dmsdos_read_cluster(struct super_block*sb,
                        unsigned char*clusterd, int clusternr,int cvfnr)
{ int ret;
       
  LOCK_READWRITE;
  
  LOG_CLUST("DMSDOS: read_cluster %d in CVF %d\n",clusternr,cvfnr+1);

  switch(dblsb[cvfnr].s_cvf_version)
  { case DBLSP:
    case DRVSP:
    case DRVSP3:
      ret=dbl_read_cluster(sb,clusterd,clusternr,cvfnr);
      break;
    case STAC3:
    case STAC4:
      ret=stac_read_cluster(sb,clusterd,clusternr,cvfnr);
      break;
    default:
      printk("DMSDOS: read_cluster: illegal cvf version flag!\n");
      ret=-EIO;
  }
  
  UNLOCK_READWRITE;
  
  return ret;
}
