#ifndef _LINUX_DMSDOS_FS_CONFIG_H
#define _LINUX_DMSDOS_FS_CONFIG_H

/*
 * The DMSDOS filesystem -- user configurable defines
 */

/**************************************************************************

   Be sure to know what you are doing if you change something herein!
   
**************************************************************************/

/* MAXFRAGMENT value:
   Increase this value if you get a 'CVF too fragmented, not mounted' message.
   To find out an appropriate value, run SCANDISK /FRAGMENT under DOS on the
   CVF dmsdos refuses to mount. It is recommended to use a slightly larger
   number here just in case you want to enlarge the CVF under DOS later. 
   See documentation for details.
   Note: For usual msdos systems 10 should be sufficient; win95 seems to
   need 20, DriveSp 3 may require 50 or higher. */
#define MAXFRAGMENT 50

/* if defined, use vmalloc instead of kmalloc */
#define USE_VMALLOC

/* this is for safety: if defined, always wait until the filesystem access
   that came first has finished before the next one is processed - undef this
   for speed improvement ...
   this shouldn't really be necessary, so it's disabled by default */
#undef SERIALIZE_RW_ACCESS

/* maximum number of clusters that can be delayed for compression -
   this is for the dmsdos compression daemon / not for write-back caching 
   A value larger than 4096 seems to slow down your system more than
   not using the daemon since the driver is busy with maintaining the cache
   most of the time. Still good performance: 1024 */
#define LISTSIZE 4096 /* must be a power of 2 */

/* time (in seconds) after which free_idle_cache returns idle buffers to
   the kernel -- which in turn decides what to do with them -- this is
   for mdfat/fat/bitfat caching */
#define MAX_CACHE_TIME 60

/* number of cvfs that can be handled at the same time */
/* the following number *must* stay below 31 */
#define MAXDBLFILES 10

/* mdfat/fat/bitfat cache sizes -- please note that the sum of these three
   are the number of fs buffers dmsdos *permanently* uses - keep small on
   systems that are short on memory */
#define MDFATCACHESIZE 40
#define DFATCACHESIZE 20
#define BITFATCACHESIZE 10

/* change the 'define' to 'undef' if you want to inhibit write access 
   by _not_ _compiling_ _in_ the low-level write code */
#define DBL_WRITEACCESS

/* undef this if you have problems with write-back cluster caching --
   then it's only write-through */
#define WRITE_BACK

/* this is for debugging only - let the bitfat check not break immediately on
   the first error, but continue until this limit is reached */
#define MAX_BITFAT_ERRORS 20

/* switch off debug logging */
#define DEFAULT_LOGLEVEL 0

/* undef this to avoid compression in general */
#define DBL_COMPRESSION

/* GUESS_HACK forces the compression scheme guessing code to return JM-0-1 as
   result when it in fact has guessed SQ-0-0. This is because SQ-0-0 is
   currently unsupported, so write access would be always uncompressed on
   SQ-0-0 compressed drives. 'undef' this to avoid this hack */ 
#define GUESS_HACK

/* default compression mode,
   you should use READ_ONLY or GUESS, further possibilities see dmsdos_fs.h */
#define DEFAULT_COMP GUESS

/* default compression level -1 */
/* 0=least efficient but fastest compression
   11=most efficient but slowest compression */
#define DEFAULT_CF   11

/* maximum number of clusters in cluster cache */
/* keep small on systems with low memory - otherwise dmsdos may eat up too
   much memory and slow down your system by causing great swapping -
   every cluster uses 8KB (32KB on drivespace 3)
   you shouldn't exceed approx. 1/4 of your physical ram
   so stay below 64 if you have only 4MB
   stay below 64 if you have only 16MB and you are using drivespace 3 
   
   Examples:
              
       Physical RAM       without drivespace 3      with drivespace 3
          2MB (*)         CCACHESIZE 32             unusable
          4MB             CCACHESIZE 64             CCACHESIZE 16
          8MB             CCACHESIZE 128            CCACHESIZE 32
       >=16MB             CCACHESIZE 256            CCACHESIZE 64
       
     (*) only with a *really small* kernel, otherwise forget it.
   
   In this define, the *upper limit* is specified, i.e. dmsdos never
   uses more than this amount for cluster caching even on extremely heavy
   disk access. However, most of the time it will just have a few clusters 
   in its cache. A value larger than 256 doesn't seem to speed up anything
   any more, but only eats memory. This value needn't be a power of 2. */
#define CCACHESIZE 64

/* time after which unused clusters are removed from the cluster cache */
#define MAX_CCACHE_TIME 240

/* whether to fill intermediate clusters on write access... supposed to be
   obsolete... define it if your files become unreadable under dos...
   and *please* send a problem report in that case */
#undef FILLWRITE

/* This defines what happens with unwritten dirty clusters that are in the 
   cache when the filesystem is unmounted. 0 means write them back, but
   compress them before. 1 means write them back without compression. 
   *** COMPRESSION TAKES SOME TIME, YOU'LL NOTICE IT *** Since most people 
   tend to press the reset button when a filesystem unmount takes some
   minutes (especially on a shutdown/reboot), I've set it to 1. So the
   unmount is fast and the users don't panic. :) */
#define UMOUNT_UCFLAG 1

/* This is for doublespace/drivespace sector allocation code. If you change
   the undef to define, allocation may be faster but may also cause a lot
   of additional fragmentation. I don't know which is better. It has only
   an effect if you really write *much* to a compressed dos partition.
   Please let me know your results if you play with this switch. */
#undef FAST_ALLOC

#endif
