/*

linux/fs/dmsdos/dmsdos_ioctl.c

DMSDOS filesystem: special functions for further dmsdos tools

******************************************************************************
DMSDOS (Doublespace/Drivespace compressed MSDOS filesystem) for Linux
written 1995,1996 by Frank Gockel

    (C) Copyright 1995,1996 by Frank Gockel

Some code of the dmsdos filesystem has been copied from the msdos filesystem
so there are the following additional copyrights:

    (C) Copyright 1992,1993 by Werner Almesberger (msdos filesystem)
    (C) Copyright 1994,1995 by Jacques Gelinas (mmap code)
    (C) Copyright 1992-1995 by Linus Torvalds

The DMSDOS filesystem was inspired by the THS filesystem (a simple doublespace
DS-0-2 compressed read-only filesystem) written 1994 by Thomas Scheuermann.

The DMSDOS filesystem is distributed under the Gnu General Public Licence.
See file COPYING for details.
******************************************************************************

*/

#include <asm/segment.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/msdos_fs.h>
#include <linux/dmsdos_fs.h>

extern Dblsb dblsb[];

int daemon_present=0;
int must_maintain_list=0;
int listcount=0;
Rwlist rwlist[LISTSIZE];
int rlist=0;
int wlist=0;

void dumpcache(void);

struct wait_queue * list_wait = NULL;
int list_lock=0;
void lock_listaccess(void)
{ while(list_lock)sleep_on(&list_wait);
  list_lock=1;
}
void unlock_listaccess(void)
{ list_lock=0;
  wake_up(&list_wait);
}

void log_statistics(void)
{ log_list_statistics();
  log_ccache_statistics();
  /*log_other_statistics();*/
}

int set_maxcluster(struct super_block*sb,int data,int cvfnr)
{ struct buffer_head*bh,*bh2;
  int i;
  unsigned char*pp;
  unsigned long int sectors;

  if(dblsb[cvfnr].s_cvf_version>=DRVSP3)
  { printk("DMSDOS: CVF %d: set_maxcluster refused - CVF is not in doublespace or drivespace<=2 format.\n",
           cvfnr+1);
    return -EINVAL;
  }
  if(data<2||data>dblsb[cvfnr].s_max_cluster2)return -EINVAL;
  /* check if higher clusters are unused */
  for(i=dblsb[cvfnr].s_max_cluster2;i>data;--i)
  { if(dbl_fat_nextcluster(sb,i,cvfnr,NULL))
    { printk("DMSDOS: CVF %d: set_maxcluster %d refused: cluster %d in use\n",
             cvfnr+1,data,i);
      return -EINVAL;
    }
  }
  
  bh=read_dbl_sector(sb,0,cvfnr);
  if(bh==NULL)return -EIO;
  bh2=read_dbl_sector(sb,dblsb[cvfnr].s_bootblock,cvfnr);
  if(bh2==NULL)
  { bh_free(sb,bh);
    return -EIO;
  }
  
  /* calculate number of sectors */
  /*sectors=(data-1)<<4;*/
  sectors=(data-1)*dblsb[cvfnr].s_sectperclust;
  sectors+=dblsb[cvfnr].s_rootdiranzentry>>4;
  pp=&(bh->b_data[14]);
  sectors+=CHS(pp);
  pp=&(bh->b_data[22]);
  sectors+=CHS(pp);
  bh->b_data[32]=sectors;
  bh->b_data[33]=sectors>>8;
  bh->b_data[34]=sectors>>16;
  bh->b_data[35]=sectors>>24;
  sectors+=CHS(pp);
  bh2->b_data[32]=sectors;
  bh2->b_data[33]=sectors>>8;
  bh2->b_data[34]=sectors>>16;
  bh2->b_data[35]=sectors>>24;
  bh_dirty(sb,bh);
  bh_dirty(sb,bh2);
  bh_free(sb,bh);
  bh_free(sb,bh2);
  
  dblsb[cvfnr].s_max_cluster=data;
  
  return 0;
}
    
int dmsdos_ioctl_dir(
	struct inode *dir,
	struct file *filp,
	unsigned int cmd,
	unsigned long data)
{ unsigned char* idata;
  int cvfnr;
  struct buffer_head*bh;
  int newval;
  int val;
  int cluster;
  int entrynr;
  int free_clusters,used_clusters,error_clusters;
  int free_sectors,used_sectors,max_size,sector,size;
  Mdfat_entry mde,dummy;
  int mdfatk,mdfatd,mdfat_sumk,mdfat_sumd;
  unsigned char buf[32];
  unsigned char *clusterd;
  int membytes;
  int length;
  int i;
  int plist;
  /*int sectors;*/
  int rawlength;

  if(dostest(dir))return -EINVAL;
  if((cvfnr=dbltest(dir))!=0)cvfnr=(cvfnr>>DMSDOS_CVFNR_BITS)-1;
  else cvfnr=(dir->i_ino>>DMSDOS_CVFNR_BITS)-1;
  idata=(unsigned char*)data;

  switch(cmd)	
  { case DMSDOS_GET_DBLSB:
      /* check dutil version number */
      if(get_fs_long(idata)<DMSDOS_LOWEST_COMPATIBLE_VERSION)
        return DMSDOS_VERSION|0x0f000000;
      memcpy_tofs(idata,(unsigned char*)&(dblsb[cvfnr]),sizeof(Dblsb));
      return DMSDOS_VERSION;
    case DMSDOS_EXTRA_STATFS:
      free_sectors=0;
      max_size=0;  
      for(sector=dblsb[cvfnr].s_datastart;sector<dblsb[cvfnr].s_dataend;
          ++sector)
      { if(dbl_bitfat_value(dir->i_sb,sector,cvfnr,NULL)==0)
        { ++free_sectors;
          size=1;
          while(dbl_bitfat_value(dir->i_sb,sector+size,cvfnr,NULL)==0)++size;
          if(size>max_size)max_size=size;
          sector+=size-1;
          free_sectors+=size-1;
        }
      }
      used_sectors=dblsb[cvfnr].s_dataend+1-dblsb[cvfnr].s_datastart
                   -free_sectors;
      mdfat_sumk=0;
      mdfat_sumd=0;
      mdfatk=0;
      mdfatd=0;
      free_clusters=0;
      used_clusters=0;
      error_clusters=0;
      for(cluster=2;cluster<=dblsb[cvfnr].s_max_cluster;++cluster)
      { if(dbl_fat_nextcluster(dir->i_sb,cluster,cvfnr,NULL)!=0)
        { dbl_mdfat_value(dir->i_sb,cluster,cvfnr,NULL,&mde);
          /*if(mdfat&0x80000000)*/
          if(mde.flags&2)
          { ++used_clusters;
            /*if(mdfat&0x40000000)++mdfatd; else ++mdfatk;*/
            if(mde.flags&1)++mdfatd; else ++mdfatk;
            mdfat_sumd+=mde.size_hi_minus_1+1;
            mdfat_sumk+=mde.size_lo_minus_1+1;
          }
          else ++error_clusters;
        }
        else ++free_clusters;
      }
      put_fs_long(free_sectors,idata);
      put_fs_long(used_sectors,idata+sizeof(long));
      put_fs_long(max_size,idata+2*sizeof(long));
      put_fs_long(free_clusters,idata+3*sizeof(long));
      put_fs_long(used_clusters,idata+4*sizeof(long));
      put_fs_long(error_clusters,idata+5*sizeof(long));
      put_fs_long(mdfat_sumk,idata+6*sizeof(long));
      put_fs_long(mdfat_sumd,idata+7*sizeof(long));
      put_fs_long(mdfatk,idata+8*sizeof(long));
      put_fs_long(mdfatd,idata+9*sizeof(long));
      return 0;
    case DMSDOS_READ_BLOCK:
      if(current->euid!=0)return -EPERM;
      sector=get_fs_long(idata);
      if(sector<0||sector>dblsb[cvfnr].s_dataend)return -EINVAL;
      bh=read_dbl_sector(dir->i_sb,sector,cvfnr);
      if(bh==NULL)return -EIO;
      memcpy_tofs(idata+sizeof(long),bh->b_data,SECTOR_SIZE);
      bh_free(dir->i_sb,bh);
      return 0;
    case DMSDOS_WRITE_BLOCK:
      if(current->euid!=0)return -EPERM;
      sector=get_fs_long(idata);
      if(sector<0||sector>dblsb[cvfnr].s_dataend)return -EINVAL;
      bh=noread_dbl_sector(dir->i_sb,sector,cvfnr);
      if(bh==NULL)return -EIO;
      memcpy_fromfs(bh->b_data,idata+sizeof(long),SECTOR_SIZE);
      bh_dirty(dir->i_sb,bh);
      bh_free(dir->i_sb,bh);
      return 0;
    case DMSDOS_READ_DIRENTRY:
      if(current->euid!=0)return -EPERM;
      cluster=get_fs_long(idata);
      if(cluster>dblsb[cvfnr].s_max_cluster)return -EINVAL;
      if(cluster<0)cluster=MSDOS_I(dir)->i_start;
      entrynr=get_fs_long(idata+sizeof(long));
      val=read_dbl_direntry(dir->i_sb,cluster,cvfnr,entrynr,buf);
      if(val<0)return val;
      memcpy_tofs(idata+2*sizeof(long),buf,32);
      return 0;
    case DMSDOS_WRITE_DIRENTRY:
      if(current->euid!=0)return -EPERM;
      cluster=get_fs_long(idata);
      if(cluster>dblsb[cvfnr].s_max_cluster)return -EINVAL;
      if(cluster<0)cluster=MSDOS_I(dir)->i_start;
      entrynr=get_fs_long(idata+sizeof(long));
      val=read_dbl_direntry(dir->i_sb,cluster,cvfnr,entrynr,buf);
      if(val<0)return val;
      bh=read_virtual_sector(dir->i_sb,(val&DMSDOS_ORIGINO_MASK)>>4,cvfnr);
      if(bh==NULL)return -EIO;
      memcpy_fromfs(&(bh->b_data[32*(val&0xf)]),idata+2*sizeof(long),32);
      bh_dirty(dir->i_sb,bh);
      bh_free(dir->i_sb,bh);
      return 0;
    case DMSDOS_READ_BITFAT:
      sector=get_fs_long(idata);
      if(sector<0||sector>dblsb[cvfnr].s_dataend)return -EINVAL;
      val=dbl_bitfat_value(dir->i_sb,sector,cvfnr,NULL);
      put_fs_long(val,idata+sizeof(long));
      return 0;
    case DMSDOS_WRITE_BITFAT:
      if(current->euid!=0)return -EPERM;
      sector=get_fs_long(idata);
      if(sector<0||sector>dblsb[cvfnr].s_dataend)return -EINVAL;
      newval=get_fs_long(idata+sizeof(long));
      dbl_bitfat_value(dir->i_sb,sector,cvfnr,&newval);
      return 0;
    case DMSDOS_READ_MDFAT:
      cluster=get_fs_long(idata);
      if(cluster>dblsb[cvfnr].s_max_cluster)return -EINVAL;
      dbl_mdfat_value(dir->i_sb,cluster,cvfnr,NULL,&mde);
      /*put_fs_long(val,idata+4);*/
      memcpy_tofs((Mdfat_entry*)get_fs_long(idata+sizeof(long)),&mde,
                  sizeof(Mdfat_entry));
      return 0;
    case DMSDOS_WRITE_MDFAT:
      if(current->euid!=0)return -EPERM;
      cluster=get_fs_long(idata);
      if(cluster>dblsb[cvfnr].s_max_cluster)return -EINVAL;
      /*newval=get_fs_long(idata+4);*/
      memcpy_fromfs(&mde,(Mdfat_entry*)get_fs_long(idata+sizeof(long)),
                    sizeof(Mdfat_entry));
      dbl_mdfat_value(dir->i_sb,cluster,cvfnr,&mde,&dummy);
      return 0;
    case DMSDOS_READ_DFAT:
      cluster=get_fs_long(idata);
      if(cluster>dblsb[cvfnr].s_max_cluster)return -EINVAL;
      val=dbl_fat_nextcluster(dir->i_sb,cluster,cvfnr,NULL);
      put_fs_long(val,idata+sizeof(long));
      return 0;
    case DMSDOS_WRITE_DFAT:
      if(current->euid!=0)return -EPERM;
      cluster=get_fs_long(idata);
      if(cluster>dblsb[cvfnr].s_max_cluster)return -EINVAL;
      newval=get_fs_long(idata+sizeof(long));
      dbl_fat_nextcluster(dir->i_sb,cluster,cvfnr,&newval);
      return 0;
    case DMSDOS_READ_CLUSTER:
      if(current->euid!=0)return -EPERM;
      cluster=get_fs_long(idata);
      if(cluster < 0 || cluster>dblsb[cvfnr].s_max_cluster)
        return -EINVAL;
      membytes=SECTOR_SIZE*dblsb[cvfnr].s_sectperclust;
      if((clusterd=(unsigned char*)MALLOC(membytes))==NULL)
      { printk("DMSDOS: ioctl: read_cluster: no memory!\n");
        return -EIO;
      }
      val=dmsdos_read_cluster(dir->i_sb,clusterd,cluster,cvfnr);
      if (val >= 0)
        memcpy_tofs(idata+sizeof(long), clusterd, membytes);
      FREE(clusterd);
      return val;
    case DMSDOS_SET_COMP:
      if(current->euid!=0)return -EPERM;
      dblsb[cvfnr].s_comp=data;
      return 0;
    case DMSDOS_SET_CF:
      if(current->euid!=0)return -EPERM;
      if(data>=12u)return -EINVAL;
      dblsb[cvfnr].s_cfaktor=data;
      return 0;
    case DMSDOS_SIMPLE_CHECK:
      val=simple_check(dir->i_sb,cvfnr,get_fs_long(idata));
      put_fs_long(val,idata);
      return 0;
    case DMSDOS_DUMPCACHE:
      dumpcache();
      return 0;
    case DMSDOS_D_ASK:
      if(current->euid!=0)return -EPERM;
      /* meldet dmsdosd an */
      daemon_present=1;
      LOG_DAEMON("DMSDOS: D_ASK\n");
      return 0;
    case DMSDOS_D_READ:
      if(current->euid!=0)return -EPERM;
      lock_listaccess();
      /*search next valid entry*/
      for(i=LISTSIZE;i>0;--i)
      { if(rwlist[rlist].flag==D_VALID)
        { /* check in mdfat that cluster is actually used */
          dbl_mdfat_value(dir->i_sb,rwlist[rlist].clusternr,rwlist[rlist].cvfnr,
                          NULL,&mde);
          if((mde.flags&3)==3)goto vr_found; /* used and uncompressed */
          rwlist[rlist].flag=D_EMPTY;  /* remove - it's garbage */
          --listcount;
          LOG_DAEMON("DMSDOS: D_READ: removing garbage entry cluster=%d cvf=%d\n",
                 rwlist[rlist].clusternr,rwlist[rlist].cvfnr+1);
        }
        rlist=(rlist+1)&(LISTSIZE-1);
      }
      unlock_listaccess();
      return 0;
   vr_found:
      cluster=rwlist[rlist].clusternr;
      cvfnr=rwlist[rlist].cvfnr;
      length=rwlist[rlist].length;
      clusterd=MALLOC(dblsb[cvfnr].s_sectperclust*SECTOR_SIZE);
      if(clusterd==NULL)
      { printk("DMSDOS: ioctl: D_READ: no memory!\n");
        unlock_listaccess();
        return 0; 
      }
      if(dmsdos_read_cluster(dir->i_sb,clusterd,cluster,cvfnr)<0)
      { printk("DMSDOS: ioctl: D_READ: read_cluster failed!\n");
        rwlist[rlist].flag=D_EMPTY;
        --listcount;
        unlock_listaccess();
        FREE(clusterd);
        return 0;   
      }
      rwlist[rlist].flag=D_IN_D_ACTION;
      put_fs_long(rlist,idata);
      put_fs_long(length,idata+sizeof(long));
      put_fs_long(rwlist[rlist].method,idata+2*sizeof(long));
      memcpy_tofs(idata+3*sizeof(long),clusterd,length);
      unlock_listaccess();
      FREE(clusterd);
      return 1;
    case DMSDOS_D_WRITE:
      if(current->euid!=0)return -EPERM;
      lock_listaccess();
      plist=get_fs_long(idata);
      if(rwlist[plist].flag==D_OVERWRITTEN){rwlist[plist].flag=D_EMPTY;--listcount;}
      if(rwlist[plist].flag!=D_IN_D_ACTION)
      { LOG_DAEMON("DMSDOS: D_WRITE: Entry not in action, flag=%d cluster=%d cvf=%d\n",
               rwlist[plist].flag,rwlist[plist].clusternr,rwlist[plist].cvfnr+1);
        unlock_listaccess();
        return 0;
      }
      
      /* will be freed in any case later, so: */
      --listcount;
      
      rawlength=get_fs_long(idata+sizeof(long));
      if(rawlength==0)
      { /* data were uncompressible */
        rwlist[plist].flag=D_EMPTY;
        unlock_listaccess();
        return 0;
      }
      /* check that cluster is used */
      dbl_mdfat_value(dir->i_sb,rwlist[plist].clusternr,rwlist[plist].cvfnr,
                      NULL,&mde);
      if((mde.flags&3)!=3)
      { rwlist[plist].flag=D_EMPTY;  /* remove - it's garbage */
        LOG_DAEMON("DMSDOS: D_WRITE: removing garbage entry cluster=%d cvf=%d\n",
               rwlist[plist].clusternr,rwlist[plist].cvfnr+1);
        unlock_listaccess();
        return 0;
      }
      clusterd=MALLOC(rawlength);
      if(clusterd==NULL)
      { printk("DMSDOS: ioctl: D_WRITE: no memory!\n");
        rwlist[plist].flag=D_EMPTY;
        unlock_listaccess();
        return 0;
      }
      length=rwlist[plist].length;
      cluster=rwlist[plist].clusternr;
      cvfnr=rwlist[plist].cvfnr;
      memcpy_fromfs(clusterd,idata+3*sizeof(long),rawlength);
      dmsdos_write_cluster(dir->i_sb,clusterd,length,cluster,0,cvfnr,
                           -rawlength); /* ucflag < 0 means raw write */
      rwlist[plist].flag=D_EMPTY;
      unlock_listaccess();
      FREE(clusterd);
      return 0;
    case DMSDOS_D_EXIT:
      if(current->euid!=0)return -EPERM;
      /* meldet dmsdosd ab */
      daemon_present=0;
      LOG_DAEMON("DMSDOS: D_EXIT\n");
      return 0;
    case DMSDOS_MOVEBACK:
      if(current->euid!=0)return -EPERM;
      return moveback(dir->i_sb,data,cvfnr);
    case DMSDOS_SET_MAXCLUSTER:
      if(current->euid!=0)return -EPERM;
      return set_maxcluster(dir->i_sb,data,cvfnr);      
    case DMSDOS_FREE_IDLE_CACHE:
      if(current->euid!=0)return -EPERM;
      free_idle_cache(dir->i_sb);
      return 0;
    case DMSDOS_SET_LOGLEVEL:
      if(current->euid!=0)return -EPERM;
      loglevel=data;
      printk("DMSDOS: ioctl: loglevel set to 0x%lx.\n",loglevel);
      return 0;
    case DMSDOS_SYNC_CCACHE:
      sync_cluster_cache(dir->i_sb);
      return 0;
    case DMSDOS_LOG_STATISTICS:
      if(current->euid!=0)return -EPERM;
      log_statistics();
      return 0;
    default:
      return -EINVAL;
  }	
	
}

int try_daemon(int clusternr, int cvfnr, int length, int method)
{ int i;

  if(daemon_present==0&&must_maintain_list==0)return 0;
  
  lock_listaccess();
  
  must_maintain_list=1;
  
  /* check list for existing entry and mark it as overwritten */
  for(i=0;i<LISTSIZE;++i)
  { if(rwlist[i].clusternr==clusternr&&rwlist[i].cvfnr==cvfnr)
    {
      if(rwlist[i].flag==D_IN_D_ACTION){rwlist[i].flag=D_OVERWRITTEN;break;}
      if(rwlist[i].flag==D_VALID){rwlist[i].flag=D_EMPTY;--listcount;break;}
    }
  }
  
  if(daemon_present==0||listcount==LISTSIZE)
  {
    if(listcount==0)must_maintain_list=0;
    unlock_listaccess();
    return 0;
  }
 
  /* find empty slot in list */
  for(i=LISTSIZE;i>0;--i)
  { if(rwlist[wlist].flag==D_EMPTY) goto w_found;
    wlist=(wlist+1)&(LISTSIZE-1);
  }
  /*this shouldn't happen - otherwise there's a count error */
  printk("DMSDOS: try_daemon: no empty slot found, listcount corrected.\n");
  listcount=LISTSIZE;
  unlock_listaccess();
  return 0;
  
 w_found:
  ++listcount;
  rwlist[wlist].clusternr=clusternr;
  rwlist[wlist].cvfnr=cvfnr;
  rwlist[wlist].length=length;
  rwlist[wlist].method=method;
  rwlist[wlist].flag=D_VALID;
  unlock_listaccess();
  return 1;
}

void clear_list_cvfnr(int cvfnr)
{ int i;

  lock_listaccess();
  for(i=0;i<LISTSIZE;++i)
  { if(rwlist[i].cvfnr==cvfnr&&rwlist[i].flag!=D_EMPTY)
    { rwlist[i].flag=D_EMPTY;
      --listcount;
    }
  }
  unlock_listaccess();
}

void clear_list(void)
{ int i;

  lock_listaccess();
  for(i=0;i<LISTSIZE;++i)rwlist[i].flag=D_EMPTY;
  listcount=0;
  unlock_listaccess();
}

void log_list_statistics()
{ int j;
  int empty;
  int valid;
  int in_action;
  int overwritten;
  
  lock_listaccess();
  
  printk("DMSDOS: list statistics:\n");
  printk("daemon_present=%d must_maintain_list=%d listcount=%d\n",
         daemon_present,must_maintain_list,listcount);
  empty=0;
  valid=0;
  in_action=0;
  overwritten=0;
    
  for(j=0;j<LISTSIZE;++j)
      { switch(rwlist[j].flag)
        { case D_EMPTY: ++empty; break;
          case D_VALID: ++valid; break;
          case D_IN_D_ACTION: ++in_action; break;
          case D_OVERWRITTEN: ++overwritten; break;
          default: printk("DMSDOS: log_list_statistics: cannot happen.\n");
        }
      }
  printk("sum: empty=%d valid=%d in_action=%d overwritten=%d\n",
         empty,valid,in_action,overwritten);
  
  unlock_listaccess();
}
