/*
linux/fs/dmsdos/dmsdos_spc.c

DMSDOS filesystem: special routines.

******************************************************************************
DMSDOS (Doublespace/Drivespace compressed MSDOS filesystem) for Linux
written 1995,1996 by Frank Gockel

    (C) Copyright 1995,1996 by Frank Gockel

Some code of the dmsdos filesystem has been copied from the msdos filesystem
so there are the following additional copyrights:

    (C) Copyright 1992,1993 by Werner Almesberger (msdos filesystem)
    (C) Copyright 1994,1995 by Jacques Gelinas (mmap code)
    (C) Copyright 1992-1995 by Linus Torvalds

The DMSDOS filesystem was inspired by the THS filesystem (a simple doublespace
DS-0-2 compressed read-only filesystem) written 1994 by Thomas Scheuermann.

The DMSDOS filesystem is distributed under the Gnu General Public Licence.
See file COPYING for details.
******************************************************************************

*/

#include <linux/sched.h>
#include <linux/ctype.h>
#include <linux/major.h>
#include <linux/blkdev.h>
#include <linux/fs.h>
#include <linux/stat.h>
#include <linux/locks.h>
#include <asm/segment.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/msdos_fs.h>
#include <linux/dmsdos_fs.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/shm.h>
#include <linux/mman.h>
#include <asm/system.h>

/* hmmm... seems to have moved in newer kernels... */
#ifdef __FOR_KERNEL_1_3_8x
#include "../fat/msbuffer.h"
#else
#include "../msdos/msbuffer.h"
#endif

#define CLPOS 26
#define DMSDOS_FAKED_BH_MAGIC 0xfff0
/* i hope this never occurs in kernel buffers... */

int dbl_cvf_inos[MAXDBLFILES];
int dbl_cvf_startcluster[MAXDBLFILES];

Dblsb dblsb[MAXDBLFILES];
unsigned long loglevel=DEFAULT_LOGLEVEL;

Acache mdfat[MDFATCACHESIZE];
Acache dfat[DFATCACHESIZE];
Acache bitfat[BITFATCACHESIZE];

int dbl_sb_dev[MAXDBLFILES];

int lowest_never_used_cvfnr=0;

struct buffer_head* read_real_sector(struct super_block* sb,int sector)
{ struct buffer_head*bh;
  LOG_LLRW("DMSDOS: read_real_sector %d\n",sector);
#ifdef __FOR_KERNEL_1_3_8x
  bh=fat_bread(sb,sector);
#else
  bh=msdos_bread(sb,sector);
#endif
  if(bh==NULL)printk("DMSDOS: read_real_sector %d failed\n",sector);
  if(bh->b_state==DMSDOS_FAKED_BH_MAGIC)
    panic("DMSDOS: This is a bug. Please stop using this version IMMEDIATELY");
  return bh;
}

struct buffer_head* noread_real_sector(struct super_block* sb,int sector)
{ struct buffer_head*bh;
  LOG_LLRW("DMSDOS: noread_real_sector %d\n",sector);
#ifdef __FOR_KERNEL_1_3_8x
  bh=fat_getblk(sb,sector);
  if(bh==NULL)printk("DMSDOS: noread_real_sector %d failed\n",sector);
  else 
  { fat_set_uptodate(sb,bh,1);
    if(bh->b_state==DMSDOS_FAKED_BH_MAGIC)
    panic("DMSDOS: This is a bug. Please stop using this version IMMEDIATELY");
  }
#else
  bh=msdos_getblk(sb,sector);
  if(bh==NULL)printk("DMSDOS: noread_real_sector %d failed\n",sector);
  else 
  { msdos_set_uptodate(sb,bh,1);
    if(bh->b_state==DMSDOS_FAKED_BH_MAGIC)
    panic("DMSDOS: This is a bug. Please stop using this version IMMEDIATELY");
  }
#endif
  return bh;
}

void bh_free(struct super_block* sb,struct buffer_head*bh)
{ if(bh)
  { if(bh->b_state==DMSDOS_FAKED_BH_MAGIC)
    { /* we free the cluster instead */
      /* a pointer to ch was saved in b_next */
      /* uh, we must check the size first.... uh no this is for bh_dirty.. */
      Cluster_head*ch=(Cluster_head*)bh->b_next;
      /*if((unsigned long)bh->b_data-(unsigned long)ch->c_data+SECTOR_SIZE>ch->c_length)
          ch->c_length=(unsigned long)bh->b_data-(unsigned long)ch->c_data+SECTOR_SIZE;*/
      ch_free(sb,ch);
      FREE(bh);
      return;
    }
  }
#ifdef __FOR_KERNEL_1_3_8x
  fat_brelse(sb,bh);
#else
  msdos_brelse(sb,bh);
#endif
}

void bh_dirty(struct super_block*sb,struct buffer_head*bh)
{ 
  if(bh)
  { if(bh->b_state==DMSDOS_FAKED_BH_MAGIC)
    { 
      /* a copy of ch was saved in b_next */
      Cluster_head*ch=(Cluster_head*)bh->b_next;
      
      /* ensure cluster is marked large enough to hold the virtual sector */
      if((unsigned long)bh->b_data-(unsigned long)ch->c_data+SECTOR_SIZE>
                                                               ch->c_length)
          ch->c_length=(unsigned long)bh->b_data-(unsigned long)ch->c_data+
                                                                 SECTOR_SIZE;

      if(ch_dirty(sb,ch,0,DIR_MAY_BE_COMPRESSED(ch->c_cvfnr) ? 0 : 1)<0)
      {
      /* now we have a serious problem here...
         ch_dirty failed... we are in danger of losing the data....
      */
         printk(KERN_EMERG "DMSDOS: Dirty virtual sector cannot be written - FILESYSTEM DAMAGE POSSIBLE! Trying to delay write.\n");
         ch_dirty_retry_until_success(sb,ch,0,
                             DIR_MAY_BE_COMPRESSED(ch->c_cvfnr) ? 0 : 1);
      }
      return; 
    }
  }
#ifdef DBL_WRITEACCESS
#ifdef __FOR_KERNEL_1_3_8x
  fat_mark_buffer_dirty(sb,bh,1);
#else
  msdos_mark_buffer_dirty(sb,bh,1); /* !!! */
#endif
#else
  printk("DMSDOS: illegal write access, ignored\n");
#endif
}

int calc_real_sector(int dblsector, int cvfnr)
{ int i;
  int sz;

  if(dblsb[cvfnr].s_first[0]==0)return dblsector;
    
  /* calculate physical sector */
  i=0;sz=dblsector;
  do
  { sz-=dblsb[cvfnr].s_anzahl[i];
    ++i;
  }
  while(sz>=0&&i<MAXFRAGMENT);
  --i;
  sz+=dblsb[cvfnr].s_anzahl[i]+dblsb[cvfnr].s_first[i];
  return sz;
}

int dos_cluster2sector(struct super_block * sb,int clusternr)
{ return (clusternr-2)*MSDOS_SB(sb)->cluster_size+MSDOS_SB(sb)->data_start;
}

struct buffer_head* read_dbl_sector(struct super_block*sb,int dblsector,
                                    int cvfnr)
{ int sector;

  LOG_LLRW("DMSDOS: read_dbl_sector: sector=%d of CVF %d\n",dblsector,cvfnr+1);
  sector=calc_real_sector(dblsector,cvfnr);
  return read_real_sector(sb,sector);
}

struct buffer_head* noread_dbl_sector(struct super_block*sb,int dblsector,
                                      int cvfnr)
{ int sector;

  LOG_LLRW("DMSDOS: noread_dbl_sector: sector=%d of CVF %d\n",dblsector,cvfnr+1);
  sector=calc_real_sector(dblsector,cvfnr);
  return noread_real_sector(sb,sector);
}

struct buffer_head* read_virtual_sector(struct super_block*sb,int vsector,
                                    int cvfnr)
{ int dbl_clust;
  int sect_offs;
  struct buffer_head*bh;
  Cluster_head*ch;
  
  /* virtual sector is the ccc...sss field of an inode */
  /* ccccccccccccccccssssss */

  LOG_LLRW("DMSDOS: read_virtual_sector: vsector=%d of CVF %d\n",vsector,cvfnr+1);
  
  dbl_clust=vsector>>6;
  sect_offs=vsector&0x3f;
  if(dbl_clust==0)return read_dbl_sector(sb,dblsb[cvfnr].s_rootdir+sect_offs,
                                         cvfnr);

  ch=ch_read(sb,dbl_clust,cvfnr);
  if(ch==NULL)
  { printk("DMSDOS: read_virtual_sector: read_cluster failed!\n");
    return NULL;
  }
  /* now setup a fake buffer_head */
  bh=MALLOC(sizeof(struct buffer_head));
  if(bh==NULL)
  { printk("DMSDOS: read_virtual_sector: no memory!\n");
    /*FREE(clusterd);*/
    ch_free(sb,ch);
    return NULL;
  }
  bh->b_state=DMSDOS_FAKED_BH_MAGIC;

  /* we must share the data */
  bh->b_data=&(ch->c_data[sect_offs*SECTOR_SIZE]);
  /* the cluster is *not* free yet, so DON'T call ch_free here */
  /* instead copy a pointer to the cluster_head somewhere so we can free it 
     later */
  bh->b_next=(struct buffer_head*)ch; /*the cast avoids compiler warning */
  return bh;
}

int setup_fragment(struct super_block*sb, int cvfnr)
{ int fragmentzaehler;
  int clusterzaehler;
  int akt_cluster;
  int folge_cluster;
  int i;
  
  LOG_REST("DMSDOS: setup_fragment CVF %d\n",cvfnr+1);
  
  /* direct ? */
  if(dbl_cvf_startcluster[cvfnr]==0)
  {  dblsb[cvfnr].s_first[0]=0;
     return 0;
  }  
  
  fragmentzaehler=0;

  folge_cluster=dbl_cvf_startcluster[cvfnr];
  
  do
  {
    clusterzaehler=0;
    dblsb[cvfnr].s_first[fragmentzaehler]=folge_cluster;
    do
    { akt_cluster=folge_cluster;
      folge_cluster=fat_access(sb,akt_cluster,-1);
      ++clusterzaehler;
    }
    while(folge_cluster==akt_cluster+1);
  
    dblsb[cvfnr].s_anzahl[fragmentzaehler]=clusterzaehler;
    LOG_REST("DMSDOS: firstclust=%d anz=%d\n",
           dblsb[cvfnr].s_first[fragmentzaehler],
           dblsb[cvfnr].s_anzahl[fragmentzaehler]);
           
    ++fragmentzaehler;  
  }
  while(folge_cluster>0&&fragmentzaehler<MAXFRAGMENT);
  if(fragmentzaehler==MAXFRAGMENT&&folge_cluster>0)
  { /* zu fragmentiert, raus */
    dbl_cvf_inos[cvfnr]=0;
    dbl_sb_dev[cvfnr]=0;
    printk("DMSDOS: CVF %d too fragmented, not mounted. See doc.\n",cvfnr+1);
    return -1; 
  }
  LOG_REST("DMSDOS: CVF %d has %d fragment(s)\n",cvfnr+1,fragmentzaehler);
  
  /* convert cluster-oriented numbers into sector-oriented ones */
  for(i=0;i<fragmentzaehler;++i)
  { /*printk("DMSDOS: umrechnen 1\n");*/
    dblsb[cvfnr].s_first[i]=dos_cluster2sector(sb,dblsb[cvfnr].s_first[i]);
    /*printk("DMSDOS: umrechnen 2\n");*/
    dblsb[cvfnr].s_anzahl[i]*=MSDOS_SB(sb)->cluster_size;
    /*printk("DMSDOS: umrechnen 3\n");*/                                          
  }
  return 0;
}

void setup_mdfat(struct super_block*sb,int cvfnr,int repair)
{ 
  struct buffer_head*bh;
  struct buffer_head*bh2;
  int i,mdfatb,fatb;
  unsigned int version_flag;
  unsigned char * pp;

    LOG_REST("DMSDOS: setup_mdfat with CVF %d\n",cvfnr+1);
    for(i=0;i<MAXFRAGMENT;++i)
    { dblsb[cvfnr].s_first[i]=0;
      dblsb[cvfnr].s_anzahl[i]=0;
    }
    if(setup_fragment(sb,cvfnr)<0)return; 
    
    LOG_REST("DMSDOS: reading %d. CVF boot block...\n",cvfnr+1);
    bh=read_dbl_sector(sb,0,cvfnr);
    if(bh==NULL)
    { printk("DMSDOS: unable to read super block of CVF %d\n",cvfnr+1);
      dbl_cvf_inos[cvfnr]=0;
      dbl_sb_dev[cvfnr]=0;
      return;
    }
    LOG_REST("DMSDOS: boot block read finished\n");
    pp=&(bh->b_data[3]);
    if(strncmp(pp,"MSDBL6.0",8)!=0&&strncmp(pp,"MSDSP6.0",8)!=0)
    { printk("DMSDOS: MSDBL/MSDSP signature not found, CVF %d skipped\n",cvfnr+1);
      dbl_cvf_inos[cvfnr]=0;
      dbl_sb_dev[cvfnr]=0;
      bh_free(sb,bh);
      return;
    }
    pp=&(bh->b_data[45]);
    dblsb[cvfnr].s_dcluster=CHS(pp);
    if(dblsb[cvfnr].s_dcluster&0x8000)dblsb[cvfnr].s_dcluster|=0xffff0000;
    pp=&(bh->b_data[36]);
    dblsb[cvfnr].s_mdfatstart=CHS(pp)+1;
    pp=&(bh->b_data[17]);
    dblsb[cvfnr].s_rootdiranzentry=CHS(pp);
    dblsb[cvfnr].s_sectperclust=((unsigned long)(bh->b_data[13]));
    pp=&(bh->b_data[39]);i=CHS(pp);/*i=res0*/
    dblsb[cvfnr].s_bootblock=i;
    pp=&(bh->b_data[14]);
    dblsb[cvfnr].s_fatstart=i+CHS(pp);
    pp=&(bh->b_data[41]);
    dblsb[cvfnr].s_rootdir=i+CHS(pp);
    pp=&(bh->b_data[43]);
    dblsb[cvfnr].s_datastart=i+CHS(pp)+2;
    pp=&(bh->b_data[57]);
    dblsb[cvfnr].s_2nd_fat_offset=0; /* dblsp doesn't have a second fat */
    dblsb[cvfnr].s_cvf_version=DBLSP;
    version_flag=bh->b_data[51];
    if(version_flag==2)dblsb[cvfnr].s_cvf_version=DRVSP;
    if(version_flag==3)dblsb[cvfnr].s_cvf_version=DRVSP3;
    if(version_flag>3)printk("DMSDOS: CVF %d strange version flag %d, assuming 0.\n",
                             cvfnr+1,version_flag);
    
      bh2=read_dbl_sector(sb,dblsb[cvfnr].s_bootblock,cvfnr);
      if(bh2==NULL)
      { printk("DMSDOS: unable to read emulated boot block of CVF %d\n",cvfnr+1);
        dbl_cvf_inos[cvfnr]=0;
        dbl_sb_dev[cvfnr]=0;
        bh_free(sb,bh);
        return;
      }
      pp=&(bh2->b_data[57]);
      if(CHL(pp)==0x20203631)dblsb[cvfnr].s_16bitfat=1;
      else if(CHL(pp)==0x20203231)dblsb[cvfnr].s_16bitfat=0;
      else if(CHL(pp)==0x20203233)
        {
          printk("DMSDOS: CVF %d has FAT32 signature, not mounted. Please report this.\n",cvfnr+1);
          dbl_cvf_inos[cvfnr]=0;
          dbl_sb_dev[cvfnr]=0;
          bh_free(sb,bh);
          bh_free(sb,bh2);
          return;
        }
      else
      {
        pp=&(bh->b_data[62]);
        dblsb[cvfnr].s_16bitfat=(CHS(pp)>32) ? 1 : 0;
        printk("DMSDOS: FAT bit size of CVF %d not recognized, guessed %d bit\n",
               cvfnr+1,CHS(pp)>32 ? 16 : 12 );
      }
      bh_free(sb,bh2); 
    
    dblsb[cvfnr].s_full=0;
    
    /* calculate maximum cluster nr (fixes lost cluster messages) */
    mdfatb=(dblsb[cvfnr].s_bootblock-dblsb[cvfnr].s_mdfatstart);
    mdfatb*=((dblsb[cvfnr].s_sectperclust>16)?102:128);
    mdfatb-=dblsb[cvfnr].s_dcluster;
    fatb=512*(dblsb[cvfnr].s_rootdir-dblsb[cvfnr].s_fatstart);
    if(dblsb[cvfnr].s_16bitfat)fatb/=2; else fatb=(2*fatb)/3;
    dblsb[cvfnr].s_max_cluster=((mdfatb<fatb)?mdfatb:fatb)-1;
    if(dblsb[cvfnr].s_16bitfat)
    { if(dblsb[cvfnr].s_max_cluster>0xFFF6)dblsb[cvfnr].s_max_cluster=0xFFF6;
    }
    else
    { if(dblsb[cvfnr].s_max_cluster>0xFF6)dblsb[cvfnr].s_max_cluster=0xFF6;
    }

    /* adapt max_cluster according to dos' limits */
    dblsb[cvfnr].s_max_cluster2=dblsb[cvfnr].s_max_cluster;
    pp=&(bh->b_data[32]);
    i=CHL(pp);
    pp=&(bh->b_data[22]);
    i-=CHS(pp);
    pp=&(bh->b_data[14]);
    i-=CHS(pp);
    i-=dblsb[cvfnr].s_rootdiranzentry>>4;
    /*i=(i>>4)+1;*/
    i=(i/dblsb[cvfnr].s_sectperclust)+1;
    if(i<=dblsb[cvfnr].s_max_cluster)
    { dblsb[cvfnr].s_max_cluster=i;
    }
    else
    { printk("DMSDOS: CVF %d: dos max_cluster=%d too large, cutting to %d.\n",
             cvfnr+1,i,dblsb[cvfnr].s_max_cluster);
    }
    
    bh_free(sb,bh);
        
    LOG_REST("DMSDOS: dcluster=%d\n",dblsb[cvfnr].s_dcluster);
    LOG_REST("DMSDOS: mdfatstart=%d\n",dblsb[cvfnr].s_mdfatstart);
    LOG_REST("DMSDOS: rootdiranzentry=%d\n",dblsb[cvfnr].s_rootdiranzentry);
    LOG_REST("DMSDOS: sectperclust=%d\n",dblsb[cvfnr].s_sectperclust);
    LOG_REST("DMSDOS: fatstart=%d\n",dblsb[cvfnr].s_fatstart);
    LOG_REST("DMSDOS: rootdir=%d\n",dblsb[cvfnr].s_rootdir);
    LOG_REST("DMSDOS: %d bit FAT\n",dblsb[cvfnr].s_16bitfat ? 16 : 12);

    /* error test */
    if(repair!=-1) /* repair==-1 means do not even check */
    {
      i=simple_check(sb,cvfnr,repair&1);
      if(i==-1||i==-2)
      { printk("DMSDOS: CVF %d has serious errors, setting to read-only.\n",
               cvfnr+1);
        dblsb[cvfnr].s_comp=READ_ONLY;
      }
      if(i==-3)
      { if(repair&2)
        { printk("DMSDOS: CVF %d has bitfat mismatches, ignored.\n",cvfnr+1);
        }
        else
        { printk("DMSDOS: CVF %d has bitfat mismatches, setting to read-only.\n",
                  cvfnr+1);
          dblsb[cvfnr].s_comp=READ_ONLY;
        }
      }
    }
    
    /* print doublespace version */
    if(dblsb[cvfnr].s_cvf_version==DBLSP&&dblsb[cvfnr].s_sectperclust==16)
    { printk("DMSDOS: CVF %d is in doublespace format (version 1).\n",cvfnr+1);
      if(dblsb[cvfnr].s_support_lfn&2)dblsb[cvfnr].s_support_lfn|=8;
    }
    else if(dblsb[cvfnr].s_cvf_version==DRVSP&&dblsb[cvfnr].s_sectperclust==16)
    { printk("DMSDOS: CVF %d is in drivespace format (version 2).\n",cvfnr+1);
      if(dblsb[cvfnr].s_support_lfn&2)dblsb[cvfnr].s_support_lfn|=8;
    }
    else if(dblsb[cvfnr].s_cvf_version==DRVSP3&&dblsb[cvfnr].s_sectperclust==64)
    { printk("DMSDOS: CVF %d is in drivespace 3 format.\n",cvfnr+1);
      if(dblsb[cvfnr].s_support_lfn&2)dblsb[cvfnr].s_support_lfn|=1;      
    }
    else
    { printk("DMSDOS: CVF %d is in unknown (new?) format, please report.\n",
             cvfnr+1);
      printk("DMSDOS: version_flag=%d sectperclust=%d\n",version_flag,
             dblsb[cvfnr].s_sectperclust);
      printk("DMSDOS: CVF %d set to read-only.\n",cvfnr+1);
      dblsb[cvfnr].s_comp=READ_ONLY;
    }
    
    dblsb[cvfnr].s_lastnear=0;
    dblsb[cvfnr].s_lastbig=0;
    dblsb[cvfnr].s_free_sectors=-1;  /* -1 means unknown */
}

void setup_stacker(struct super_block*sb,int cvfnr,int repair)
{ 
  struct buffer_head*bh;
  struct buffer_head*bh2;
  int i;
  unsigned char * pp, *p;
  unsigned char buf[512];
  unsigned char b,c;
  int SectSize, ClustSects, ClustSize, ReservSects, FATCnt;
  int RootDirEnt, TotalSects, FATSize, HidenSects, FirstRootSect;
  int FirstDataSect, FirstDataSect2, FAT12, FirstFATSect;
  int StacVersion;
  /* parameters of virtual DOS drive */
  int BB_FirstDataSect, BB_ClustCnt, BB_SectSize, BB_TotalSects;

    printk("DMSDOS: setup_stacker with CVF %d\n",cvfnr+1);
    for(i=0;i<MAXFRAGMENT;++i)
    { dblsb[cvfnr].s_first[i]=0;
      dblsb[cvfnr].s_anzahl[i]=0;
    }
    if(setup_fragment(sb,cvfnr)<0)return; 
    
    LOG_REST("DMSDOS: reading %d. CVF super block...\n",cvfnr+1);
    bh=read_dbl_sector(sb,0,cvfnr);
    if(bh==NULL)
    { printk("DMSDOS: unable to read super block of CVF %d\n",cvfnr+1);
      dbl_cvf_inos[cvfnr]=0;
      dbl_sb_dev[cvfnr]=0;
      return;
    }
    LOG_REST("DMSDOS: super block read finished\n");
    pp=&(bh->b_data[0]);
    if(strncmp(pp,"STACKER",7)!=0)
    { printk("DMSDOS: STACKER signature not found, CVF %d skipped\n",cvfnr+1);
      dbl_cvf_inos[cvfnr]=0;
      dbl_sb_dev[cvfnr]=0;
      bh_free(sb,bh);
      return;
    }
    /* copy block (must not directly modify kernel buffer!!!) */
    memcpy(buf,bh->b_data,SECTOR_SIZE);
    /* decode super block */
    for(i=0x30,p=buf+0x50,b=buf[0x4c];i--;p++)
    { b=0xc4-b;
      b=b<0x80?b*2:b*2+1;
      b^=c=*p;
      *p=b;b=c;
    }
    if(buf[0x4e]!=0xa||buf[0x4f]!=0x1a)
    { printk("DMSDOS: Stacker 0x1A0A signature not found, CVF %d skipped.\n",
             cvfnr+1);
      dbl_cvf_inos[cvfnr]=0;
      dbl_sb_dev[cvfnr]=0;
      bh_free(sb,bh);
      return;                           
    }
    /* extract important info */
    pp=&(buf[0x6C]);
    TotalSects=CHL(pp);
    pp=&(buf[0x70]);
    dblsb[cvfnr].s_bootblock=CHS(pp);
    pp=&(buf[0x74]);
    dblsb[cvfnr].s_mdfatstart=CHS(pp); /* here it's AMAP start !!! */
    pp=&(buf[0x76]);
    FirstFATSect=dblsb[cvfnr].s_fatstart=CHS(pp);
    pp=&(buf[0x7a]);
    FirstDataSect2=dblsb[cvfnr].s_datastart=CHS(pp);
    pp=&(buf[0x60]);
    StacVersion=CHS(pp);
    if(StacVersion>=410)dblsb[cvfnr].s_cvf_version=STAC4;
    else dblsb[cvfnr].s_cvf_version=STAC3;

    /* if(buf[0x64]==9)dblsb[cvfnr].s_cvf_version=STAC4;
    else dblsb[cvfnr].s_cvf_version=STAC3; */
    
    /* now we need the boot block */    
    bh2=read_dbl_sector(sb,dblsb[cvfnr].s_bootblock,cvfnr);
    if(bh2==NULL)
    { printk("DMSDOS: unable to read emulated boot block of CVF %d\n",cvfnr+1);
      dbl_cvf_inos[cvfnr]=0;
      dbl_sb_dev[cvfnr]=0;
      bh_free(sb,bh);
      return;
    }
    /* read values */
    dblsb[cvfnr].s_sectperclust=bh2->b_data[0xd];
    pp=&(bh2->b_data[0x11]);
    dblsb[cvfnr].s_rootdiranzentry=CHS(pp);
    
    pp=&(buf[0x62]);         SectSize=CHS(pp);
    pp=&(bh2->b_data[0xB]);  BB_SectSize=CHS(pp);
    if(SectSize!=SECTOR_SIZE||BB_SectSize!=SECTOR_SIZE)
      printk("DMSDOS: Stacker sector size not 512 bytes, hmm...\n");
    ClustSects=bh2->b_data[0xD];
    ClustSize=ClustSects*SectSize;
    pp=&(bh2->b_data[0xE]);  ReservSects=CHS(pp);
    FATCnt=bh2->b_data[0x10];
    pp=&(bh2->b_data[0x11]); RootDirEnt=CHS(pp);
    pp=&(bh2->b_data[0x13]); BB_TotalSects=CHS(pp);
    if(!BB_TotalSects)
      { pp=&(bh2->b_data[0x20]); BB_TotalSects=CHL(pp);};
    pp=&(bh2->b_data[0x16]); FATSize=CHS(pp);
    pp=&(bh2->b_data[0x1B]); HidenSects=CHS(pp);
    if(BB_SectSize!=SectSize)printk("DMSDOS: Inconsistent sector length\n");
    FirstRootSect=FirstFATSect+3*FATCnt*FATSize;
    
    dblsb[cvfnr].s_2nd_fat_offset=3*(FATCnt-1)*FATSize;
    
    /* Number of sectors in root directory */
    FirstDataSect=((long)RootDirEnt*0x20+SectSize-1)/SectSize;
    /* Emulated data start sector for DOS */
    BB_FirstDataSect=FirstDataSect+FATCnt*FATSize+ReservSects;
      /* ??? +HidenSects; */
    /* Real data start sector */
    FirstDataSect+=FirstRootSect;
    /* Counting BB_ClustCnt from emulated boot block */
    BB_ClustCnt=(BB_TotalSects-BB_FirstDataSect)/ClustSects+1;
    if(BB_ClustCnt>=0xFEE)FAT12=0; else FAT12=1;
    if(FirstDataSect2!=FirstDataSect)
    { printk("DMSDOS: Inconsistent first data sector number. Mounting READ ONLY.\n");
      printk("In header found %u but computed %u\n",(unsigned)FirstDataSect2,(unsigned)FirstDataSect);
      dblsb[cvfnr].s_comp=READ_ONLY;
    } 
            
    printk("DMSDOS: Stac version %u start of FAT %u, root %u, data %u; FATSize %u; FATCnt %u; clusts %u; sects %u\n",
     (unsigned)StacVersion,(unsigned)FirstFATSect,(unsigned)FirstRootSect,
     (unsigned)FirstDataSect,(unsigned)FATSize,(unsigned)FATCnt,
     (unsigned)BB_ClustCnt,(unsigned)BB_TotalSects);
    
    /* try dos standard method to detect fat bit size */
    pp=&(bh2->b_data[57]);
    if(CHL(pp)==0x20203631)dblsb[cvfnr].s_16bitfat=1;
    else if(CHL(pp)==0x20203231)dblsb[cvfnr].s_16bitfat=0;
    else /* failed, guess it */
      {
        dblsb[cvfnr].s_16bitfat=1-FAT12;
        printk("DMSDOS: FAT bit size of CVF %d not recognized, guessed %d bit\n",
               cvfnr+1,(1-FAT12) ? 16 : 12 );
      }
    /* checking that clusters fits in FAT */
    if(BB_ClustCnt>(FAT12?(SECTOR_SIZE*FATSize*2)/3:(SECTOR_SIZE*FATSize)/2))
    { printk("DMSDOS: FAT size does not match cluster count of CVF %d. Mounting READ ONLY.\n",
        cvfnr+1);
      dblsb[cvfnr].s_comp=READ_ONLY;
    };

    bh_free(sb,bh2); 
    dblsb[cvfnr].s_full=0;
    bh_free(sb,bh);
    
    dblsb[cvfnr].s_rootdir=FirstRootSect;
    dblsb[cvfnr].s_max_cluster=dblsb[cvfnr].s_max_cluster2=BB_ClustCnt;

    LOG_REST("DMSDOS: mdfatstart=%d\n",dblsb[cvfnr].s_mdfatstart);
    LOG_REST("DMSDOS: rootdiranzentry=%d\n",dblsb[cvfnr].s_rootdiranzentry);
    LOG_REST("DMSDOS: sectperclust=%d\n",dblsb[cvfnr].s_sectperclust);
    LOG_REST("DMSDOS: fatstart=%d\n",dblsb[cvfnr].s_fatstart);
    LOG_REST("DMSDOS: rootdir=%d\n",dblsb[cvfnr].s_rootdir);
    LOG_REST("DMSDOS: %d bit FAT\n",dblsb[cvfnr].s_16bitfat ? 16 : 12);
 
    /* error test */
    if(repair!=-1) /* repair==-1 means do not even check */
    {
      i=simple_check(sb,cvfnr,repair&1);
      if(i==-1||i==-2)
      { printk("DMSDOS: CVF %d has serious errors, setting to read-only.\n",
               cvfnr+1);
        dblsb[cvfnr].s_comp=READ_ONLY;
      }
      if(i==-3)
      { if(repair&2)
        { printk("DMSDOS: CVF %d has bitfat mismatches, ignored.\n",cvfnr+1);
        }
        else
        { printk("DMSDOS: CVF %d has bitfat mismatches, setting to read-only.\n",
                  cvfnr+1);
          dblsb[cvfnr].s_comp=READ_ONLY;
        }
      }
    }

    dblsb[cvfnr].s_lastnear=0;
    dblsb[cvfnr].s_lastbig=0;
    dblsb[cvfnr].s_free_sectors=-1;  /* -1 means unknown */
}


struct wait_queue * mdfatwait=NULL;
int mdfatlock=0;
void lock_mdfat(void)
{ while(mdfatlock)sleep_on(&mdfatwait);
  mdfatlock=1;
}
void unlock_mdfat(void)
{ mdfatlock=0;
  wake_up(&mdfatwait);
}

struct wait_queue * dfatwait=NULL;
int dfatlock=0;
void lock_dfat(void)
{ while(dfatlock)sleep_on(&dfatwait);
  dfatlock=1;
}
void unlock_dfat(void)
{ dfatlock=0;
  wake_up(&dfatwait);
}

struct wait_queue * bitfatwait=NULL;
int bitfatlock=0;
void lock_bitfat(void)
{ while(bitfatlock)sleep_on(&bitfatwait);
  bitfatlock=1;
}
void unlock_bitfat(void)
{ bitfatlock=0;
  wake_up(&bitfatwait);
}

void read_mdfat_sector(struct super_block*sb,int area,int merk_i,int cvfnr)
{
  LOG_MDFAT("DMSDOS: reading dbl_mdfat sector %d of CVF %d\n",area,cvfnr+1);
  
  if(mdfat[merk_i].a_buffer!=NULL)bh_free(sb,mdfat[merk_i].a_buffer);
  if((mdfat[merk_i].a_buffer=read_dbl_sector(sb,area,cvfnr))==NULL)
  { panic("DMSDOS: unable to read mdfat\n");
  }
  mdfat[merk_i].a_area=area;
  mdfat[merk_i].a_time=CURRENT_TIME;
  mdfat[merk_i].a_acc=0;
  mdfat[merk_i].a_cvfnr=cvfnr;
}

void read_dfat_sector(struct super_block*sb,int area,int merk_i,int cvfnr)
{
  LOG_DFAT("DMSDOS: reading dbl_fat sector %d of CVF %d\n",area,cvfnr+1);
  
  if(dfat[merk_i].a_buffer!=NULL)bh_free(sb,dfat[merk_i].a_buffer);
  if((dfat[merk_i].a_buffer=read_dbl_sector(sb,area,cvfnr))==NULL)
  { panic("DMSDOS: unable to read dfat\n");
  }
  dfat[merk_i].a_area=area;
  dfat[merk_i].a_time=CURRENT_TIME;
  dfat[merk_i].a_acc=0;
  dfat[merk_i].a_cvfnr=cvfnr;
  
  /*printk("cluster 2 fat entry=%d\n",mdfat[cvfnr].fat_data[merk_i][2]);*/ 
}

void read_bitfat_sector(struct super_block*sb,int area,int merk_i,int cvfnr)
{
  LOG_BITFAT("DMSDOS: reading bitfat sector %d of CVF %d\n",area,cvfnr+1);
  
  if(bitfat[merk_i].a_buffer!=NULL)bh_free(sb,bitfat[merk_i].a_buffer);
  if((bitfat[merk_i].a_buffer=read_dbl_sector(sb,area,cvfnr))==NULL)
  { panic("DMSDOS: unable to read bitfat\n");
  }
  bitfat[merk_i].a_area=area;
  bitfat[merk_i].a_time=CURRENT_TIME;
  bitfat[merk_i].a_acc=0;
  bitfat[merk_i].a_cvfnr=cvfnr;
}

void u_dumpcache(Acache*c)
{ printk("area=%d time=%ld acc=%d cvfnr=%d buffer=%p\n",c->a_area,c->a_time,
         c->a_acc,c->a_cvfnr+1,c->a_buffer);
}

void dumpcache(void)
{ int i;
  
  printk("DMSDOS: mdfat cache:\n");
  for(i=0;i<MDFATCACHESIZE;++i)u_dumpcache(&(mdfat[i]));
  printk("DMSDOS: dfat cache:\n");
  for(i=0;i<DFATCACHESIZE;++i)u_dumpcache(&(dfat[i]));
  printk("DMSDOS: bitfat cache:\n");
  for(i=0;i<BITFATCACHESIZE;++i)u_dumpcache(&(bitfat[i]));
}

void u_free_idle_cache(struct super_block*sb,Acache*c)
{ if(c->a_buffer!=NULL&&c->a_time-CURRENT_TIME>MAX_CACHE_TIME)
  { bh_free(sb,c->a_buffer);
    c->a_buffer=NULL;
    c->a_time=0;
    c->a_acc=0;
  }
}

void free_idle_cache(struct super_block*sb)
{ int i;
  
  for(i=0;i<MDFATCACHESIZE;++i)u_free_idle_cache(sb,&(mdfat[i]));
  for(i=0;i<DFATCACHESIZE;++i)u_free_idle_cache(sb,&(dfat[i]));
  for(i=0;i<BITFATCACHESIZE;++i)u_free_idle_cache(sb,&(bitfat[i]));
  
  /* handle cluster cache */
  free_idle_ccache(sb);
}

void dbl_mdfat_value(struct super_block*sb,int clusternr,int cvfnr,
                     Mdfat_entry*new,
                     Mdfat_entry*mde)
{ int i;
  int area;
  int pos;
  int offset;
  unsigned long min_time;
  unsigned int min_acc;
  int merk_i,merk_i2,merk_i3;
  unsigned char * pp;
  int res;
  unsigned char mdfat_raw_field[5];

  if(dblsb[cvfnr].s_cvf_version>=STAC3)
  { /* stacker */ 
    if(dblsb[cvfnr].s_16bitfat)pos=clusternr*4;
    else pos=clusternr*3;
    area=pos/SECTOR_SIZE;
    offset=pos%SECTOR_SIZE;
    area=(area/6)*9+(area%6)+3+dblsb[cvfnr].s_fatstart; /* yes!!! */
  }

  else
  { /* dblspace/drvspace */
    if(dblsb[cvfnr].s_sectperclust>16)
      /* drvspace 3 */
      pos=(dblsb[cvfnr].s_dcluster+clusternr)*5
          +((dblsb[cvfnr].s_dcluster+clusternr)/102)*2
          +512*dblsb[cvfnr].s_mdfatstart;
    else  
      pos=(dblsb[cvfnr].s_dcluster+clusternr)*4+512*dblsb[cvfnr].s_mdfatstart;
    area=pos/SECTOR_SIZE;
    offset=(pos%SECTOR_SIZE);
  }                           
  
  lock_mdfat();
  
  merk_i2=-1; /* never free this cached sector */
  
mdfat_redo:
  
  min_time=mdfat[0].a_time;
  min_acc=mdfat[0].a_acc;
  merk_i=0;
  if(merk_i2==0)
  { min_time=mdfat[1].a_time;
    min_acc=mdfat[1].a_acc;
    merk_i=1;
  }
  /* find area and cvfnr in cache */
  for(i=0;i<MDFATCACHESIZE;++i)
  { if( ( mdfat[i].a_time<min_time||
            (mdfat[i].a_time==min_time&&mdfat[i].a_acc<min_acc)
        ) &&merk_i2!=i)    
    { min_time=mdfat[i].a_time;
      min_acc=mdfat[i].a_acc;
      merk_i=i;
    }
    if(mdfat[i].a_buffer!=NULL&&area==mdfat[i].a_area&&cvfnr==mdfat[i].a_cvfnr)
    { /* found */
      if(mdfat[i].a_time==CURRENT_TIME)++mdfat[i].a_acc;
      else
      { mdfat[i].a_time=CURRENT_TIME;
        mdfat[i].a_acc=0;
      }
      merk_i=i;
      goto mdfat_okay;      /************* goto ************************/
    }
  }
  /* merk_i = least recently used entry number */
  read_mdfat_sector(sb,area,merk_i,cvfnr);

mdfat_okay:
/*
  pp=&(mdfat[merk_i].a_buffer->b_data[offset]);
  res=CHL(pp);
  if(new)
  { pp[0]=*new;pp[1]=*new>>8;pp[2]=*new>>16;pp[3]=*new>>24;
    bh_dirty(sb,mdfat[merk_i].a_buffer);
    /printk("DMSDOS: mdfat write cluster=%d value=0x%08x\n",clusternr,*new);/
  }
  unlock_mdfat();
  return res;
*/

  if(dblsb[cvfnr].s_cvf_version>=STAC3)
  { /* stacker */
    int anzb=3;
    if(dblsb[cvfnr].s_16bitfat)anzb=4;
  
    /* read 2. sector if necessary */
    if(offset+anzb-1>511&&merk_i2==-1)
    { merk_i2=merk_i;
      ++area;
      goto mdfat_redo;
    }
    if(offset+anzb-1>511)
    { merk_i3=merk_i;
      merk_i=merk_i2;
      merk_i2=merk_i3;
    }
    else merk_i2=merk_i;
  
    /* copy data in mdfat_raw_field (anzb byte) */
    mdfat_raw_field[3]=0; /* in case only 3 bytes to read */
    for(i=0;i<anzb;++i)
    { if(i+offset<512)
        mdfat_raw_field[i]=mdfat[merk_i].a_buffer->b_data[i+offset];
      else
        mdfat_raw_field[i]=mdfat[merk_i2].a_buffer->b_data[i+offset-512];
    }
  
    /* setup mde */
    mde->sector_minus_1=CHS(mdfat_raw_field)+((mdfat_raw_field[3]&0x3f)<<16)-1;
    mde->unknown=(mdfat_raw_field[2])&0xf0; /* same as flags here */
    mde->size_lo_minus_1=(mdfat_raw_field[2]&0xf)+((mdfat_raw_field[3]>>2)&0x30);
    mde->size_hi_minus_1=mde->size_lo_minus_1;  /* for compatibility */
    mde->flags=(mdfat_raw_field[2])&0xf0;   /* unshifted like in sd4_c.cc */
    /* set used and compressed bits in flags (for compatibility) */
    if(mde->sector_minus_1+1)mde->flags|=2;
    switch(mde->flags&0xa0)
    { case 0x80:
      case 0x00:
        if(mde->size_lo_minus_1+1==dblsb[cvfnr].s_sectperclust)
          mde->flags|=1;
        break;
      case 0x20:
        mde->flags|=1;
        break;
      default:
        /* correct length for compatibility */
        mde->size_hi_minus_1=dblsb[cvfnr].s_sectperclust-1;
    }
    
    LOG_MDFAT("DMSDOS: dbl_mdfat_value : cvfnr %u cluster %u\n",
           (unsigned)cvfnr,(unsigned)clusternr);
    LOG_MDFAT("DMSDOS:    sector %u len %u flags 0x%X raw 0x%02X 0x%02X 0x%02X 0x%02X\n",
           (unsigned)(mde->sector_minus_1+1),(unsigned)(mde->size_lo_minus_1+1),
           (unsigned)mde->flags,
           (unsigned)(mdfat_raw_field[0]),(unsigned)(mdfat_raw_field[1]),
           (unsigned)(mdfat_raw_field[2]),(unsigned)(mdfat_raw_field[3]));
    LOG_MDFAT("DMSDOS:    pos %u area %u offset %u\n",
           (unsigned)pos,(unsigned)area,(unsigned)offset);
  
    /* if new!=NULL, setup anzb byte from new in mdfat_raw_field */
    if(new)
    { mdfat_raw_field[0]=(new->sector_minus_1+1);
      /* unknown bits ignored */
      mdfat_raw_field[1]=(new->sector_minus_1+1)>>8;
      mdfat_raw_field[3]=((new->sector_minus_1+1)>>16)&0x3f;
      mdfat_raw_field[3]|=(new->size_lo_minus_1<<2)&0xc0;
      mdfat_raw_field[2]=new->size_lo_minus_1&0x0f;
      mdfat_raw_field[2]|=new->flags&0xf0; /* unshifted like in sd4_c.cc */
    
      /* write back mdfat_raw_entry */
      for(i=0;i<anzb;++i)
      { if(i+offset<512)
          mdfat[merk_i].a_buffer->b_data[i+offset]=mdfat_raw_field[i];
        else
          mdfat[merk_i2].a_buffer->b_data[i+offset-512]=mdfat_raw_field[i];
      }
    
      /* mark buffer dirty (both if necessary) */
      bh_dirty(sb,mdfat[merk_i].a_buffer);
      if(merk_i!=merk_i2)bh_dirty(sb,mdfat[merk_i2].a_buffer);
      /* write second mdfat if it exists */
      if(dblsb[cvfnr].s_2nd_fat_offset)
      { struct buffer_head* bh;
      
        bh=noread_dbl_sector(sb,
                mdfat[merk_i].a_area+dblsb[cvfnr].s_2nd_fat_offset,cvfnr);
        if(bh==NULL)panic("unable to read second mdfat");
        memcpy(bh->b_data,mdfat[merk_i].a_buffer->b_data,SECTOR_SIZE);
        bh_dirty(sb,bh);
        bh_free(sb,bh);
        if(merk_i!=merk_i2)
        { bh=noread_dbl_sector(sb,
                 mdfat[merk_i2].a_area+dblsb[cvfnr].s_2nd_fat_offset,cvfnr);
          if(bh==NULL)panic("unable to read second mdfat");
          memcpy(bh->b_data,mdfat[merk_i2].a_buffer->b_data,SECTOR_SIZE);
          bh_dirty(sb,bh);
          bh_free(sb,bh);                              
        }
      }
    
    }
  
    unlock_mdfat();
    return;
  }

  if(dblsb[cvfnr].s_sectperclust<=16)
  { /* normal doublesp/drivesp */
    pp=&(mdfat[merk_i].a_buffer->b_data[offset]);
    res=CHL(pp);
    mde->sector_minus_1=res&0x1fffff;
    mde->unknown=(res&0x00200000)>>21;
    mde->size_lo_minus_1=(res&0x03c00000)>>22;
    mde->size_hi_minus_1=(res&0x3c000000)>>26;
    mde->flags=((res&0xC0000000)>>30)&3;
    if(new)
    { res=new->sector_minus_1;res&=0x1fffff;
      /* unknown bit ??? don't know... set to zero */
      res|=new->size_lo_minus_1<<22;res&=0x03ffffff;
      res|=new->size_hi_minus_1<<26;res&=0x3fffffff;
      res|=new->flags<<30;
      pp[0]=res;pp[1]=res>>8;pp[2]=res>>16;pp[3]=res>>24;
      bh_dirty(sb,mdfat[merk_i].a_buffer);
    }
    unlock_mdfat();
    return;
  }

  /* drivspace 3 mdfat access  _experimental_ but seems to work */

  /* read 2. sector if necessary */
  if(offset+4>511&&merk_i2==-1)
  { merk_i2=merk_i;
    ++area;
    goto mdfat_redo;
  }
  if(offset+4>511)
  { merk_i3=merk_i;
    merk_i=merk_i2;
    merk_i2=merk_i3;
  }
  else merk_i2=merk_i;
  
  /* copy data in mdfat_raw_field (5 byte) */
  for(i=0;i<5;++i)
  { if(i+offset<512)
      mdfat_raw_field[i]=mdfat[merk_i].a_buffer->b_data[i+offset];
    else
      mdfat_raw_field[i]=mdfat[merk_i2].a_buffer->b_data[i+offset-512];
  }
  
  /* setup mde */
  mde->sector_minus_1=CHL(mdfat_raw_field)&0xffffff;
  mde->unknown=(CHL(mdfat_raw_field)&0x3000000)>>24;
  mde->size_lo_minus_1=(mdfat_raw_field[3]>>2)&0x3f;
  mde->size_hi_minus_1=mdfat_raw_field[4]&0x3f;
  mde->flags=(mdfat_raw_field[4]>>6)&3;
  
  /* if new!=NULL, setup 5 byte from new in mdfat_raw_field */
  if(new)
  { mdfat_raw_field[0]=new->sector_minus_1&0xffffff;
    /* unknown bits set to zero */
    mdfat_raw_field[1]=new->sector_minus_1>>8;
    mdfat_raw_field[2]=new->sector_minus_1>>16;
    mdfat_raw_field[3]=(new->sector_minus_1>>24)&3;
    mdfat_raw_field[3]|=new->size_lo_minus_1<<2;
    mdfat_raw_field[4]=new->size_hi_minus_1&0x3f;
    mdfat_raw_field[4]|=new->flags<<6;
    
    /* write back mdfat_raw_entry */
    for(i=0;i<5;++i)
    { if(i+offset<512)
        mdfat[merk_i].a_buffer->b_data[i+offset]=mdfat_raw_field[i];
      else
        mdfat[merk_i2].a_buffer->b_data[i+offset-512]=mdfat_raw_field[i];
    }
    
    /* mark buffer dirty (both if necessary) */
    bh_dirty(sb,mdfat[merk_i].a_buffer);
    if(merk_i!=merk_i2)bh_dirty(sb,mdfat[merk_i2].a_buffer);
    
  }
  
  unlock_mdfat();
  return;
}

int dbl_mdfat_cluster2sector(struct super_block*sb,int clusternr,int cvfnr)
{ Mdfat_entry mde;

  if(clusternr==0)return dblsb[cvfnr].s_rootdir;
  dbl_mdfat_value(sb,clusternr,cvfnr,NULL,&mde);
  return mde.sector_minus_1+1;
}

int dbl_fat_nextcluster(struct super_block*sb,int clusternr,int cvfnr,int*new)
{ int i;
  int area;
  int pos;
  int offset;
  unsigned long min_time;
  unsigned int min_acc;
  int merk_i;
  int res;
  int newval;
  int merk_i2;
  int merk_i3;
  int offset2;
  
  merk_i2=-1; /* never free this cached sector */

/*
  pos= dblsb[cvfnr].s_fatstart*512;
  pos+= dblsb[cvfnr].s_16bitfat ? clusternr*2 : clusternr*3/2;
       
  area=pos/SECTOR_SIZE;
  offset=pos%SECTOR_SIZE;                           
*/

  pos= dblsb[cvfnr].s_16bitfat ? clusternr*2 : clusternr*3/2;
  area=pos/SECTOR_SIZE;
  offset=pos%SECTOR_SIZE;
  /* adapt area for stacker */
  if(dblsb[cvfnr].s_cvf_version>=STAC3)area=(area/3)*9+(area%3);
  
  area+=dblsb[cvfnr].s_fatstart;
  
  lock_dfat();
  
dfat_retry:
  
  min_time=dfat[0].a_time;
  min_acc=dfat[0].a_acc;
  merk_i=0;
  if(merk_i2==0)
  { min_time=dfat[1].a_time;
    min_acc=dfat[1].a_acc;
    merk_i=1;
  }
  /* find area and cvfnr in cache */
  for(i=0;i<DFATCACHESIZE;++i)
  { if( ( dfat[i].a_time<min_time ||
           (dfat[i].a_time==min_time&&dfat[i].a_acc<min_acc) 
        ) &&merk_i2!=i)
    { min_time=dfat[i].a_time;
      min_acc=dfat[i].a_acc;
      merk_i=i;
    }
    if(dfat[i].a_buffer!=NULL&&area==dfat[i].a_area&&cvfnr==dfat[i].a_cvfnr)
    { /* found */
      if(dfat[i].a_time==CURRENT_TIME)++dfat[i].a_acc;
      else
      { dfat[i].a_time=CURRENT_TIME;
        dfat[i].a_acc=0;
      }
      merk_i=i;
      goto dfat_okay;
    }
  }
  /* merk_i = least recently used entry */
  read_dfat_sector(sb,area,merk_i,cvfnr);
  
dfat_okay:
  if(offset==511&&merk_i2==-1)
  { /* bertrag, mist, folgesektor auch noch lesen, alten nicht rauswerfen */
    merk_i2=merk_i;
    ++area;
    goto dfat_retry;
  }
  if(offset==511)
  { /* gerade zweiten sektor gelesen */
    merk_i3=merk_i;
    merk_i=merk_i2; /* wiederherstellen */
    merk_i2=merk_i3; /* folgesektor */
    offset2=0;
  }
  else
  { /* normal */
    merk_i2=merk_i;
    offset2=offset+1;
  }

LOG_DFAT("DMSDOS FAT lookup: area=%d merk_i=%d merk_i2=%d offset=%d offset2=%d\n",
       area,merk_i,merk_i2,offset,offset2);
LOG_DFAT("DMSDOS FAT lookup: low=%d high=%d\n",
       dfat[merk_i].a_buffer->b_data[offset],
       dfat[merk_i2].a_buffer->b_data[offset2]); 

  res=dfat[merk_i].a_buffer->b_data[offset];
  res&=0xff; /* grmpf... sign problems !!!! this is brutal but it works */
  res|=dfat[merk_i2].a_buffer->b_data[offset2]<<8;
  res&=0xffff;
  
  if(new)
  { if(dblsb[cvfnr].s_16bitfat) newval=*new&0xFFFF;
    else
    { if(clusternr&1) newval=(res&0xF) | ((*new&0xFFF)<<4);
      else newval=(res&0xF000) | (*new&0xFFF);
    }
    dfat[merk_i].a_buffer->b_data[offset]=newval;
    dfat[merk_i2].a_buffer->b_data[offset2]=newval>>8;
    bh_dirty(sb,dfat[merk_i].a_buffer);
    if(merk_i!=merk_i2)bh_dirty(sb,dfat[merk_i2].a_buffer);
    /* write second fat if it exists */
      if(dblsb[cvfnr].s_2nd_fat_offset)
      { struct buffer_head* bh;
      
        bh=noread_dbl_sector(sb,
               dfat[merk_i].a_area+dblsb[cvfnr].s_2nd_fat_offset,cvfnr);
        if(bh==NULL)panic("unable to read second fat");
        memcpy(bh->b_data,dfat[merk_i].a_buffer->b_data,SECTOR_SIZE);
        bh_dirty(sb,bh);
        bh_free(sb,bh);
        if(merk_i!=merk_i2)
        { bh=noread_dbl_sector(sb,
               dfat[merk_i2].a_area+dblsb[cvfnr].s_2nd_fat_offset,cvfnr);
          if(bh==NULL)panic("unable to read second fat");
          memcpy(bh->b_data,dfat[merk_i2].a_buffer->b_data,SECTOR_SIZE);
          bh_dirty(sb,bh);
          bh_free(sb,bh);                              
        }
      }

  }
  unlock_dfat();
  if(dblsb[cvfnr].s_16bitfat)return res>=0xFFF7 ? -1 : res;
  if(clusternr&1)res>>=4; else res&=0xfff;
  return res>=0xff7 ? -1 : res;
}

int dbl_bitfat_value(struct super_block*sb,int sektornr,int cvfnr,int*new)
{ int i;
  int area;
  int pos;
  int offset;
  unsigned long min_time;
  unsigned int min_acc;
  int merk_i;
  unsigned char * pp;
  int bitmask,shiftval;
  int res;
  int newval;
  
  if(sektornr<dblsb[cvfnr].s_datastart)return -1;
  if(sektornr>dblsb[cvfnr].s_dataend)return -1;
  if(dblsb[cvfnr].s_cvf_version==STAC3)
  { /* stacker 3 */
    pos=((sektornr-dblsb[cvfnr].s_datastart)>>3);
    offset=pos%SECTOR_SIZE;
    area=pos/SECTOR_SIZE+dblsb[cvfnr].s_mdfatstart; /* misused */
    shiftval=((sektornr-dblsb[cvfnr].s_datastart)&7);
    bitmask=0x1;
  } /* stacker 4 */
  else if(dblsb[cvfnr].s_cvf_version==STAC4)
  { pos=((sektornr-dblsb[cvfnr].s_datastart)>>2);
    offset=pos%SECTOR_SIZE;
    area=pos/SECTOR_SIZE+dblsb[cvfnr].s_mdfatstart; /* misused */
    shiftval=((sektornr-dblsb[cvfnr].s_datastart)&3)*2;
    bitmask=0x3;
  }
  else
  { /* dblspace/drvspace */
    pos=SECTOR_SIZE+((sektornr-dblsb[cvfnr].s_datastart)>>4)*2;
    area=pos/SECTOR_SIZE;
    offset=(pos%SECTOR_SIZE);
    bitmask=0x8000>>((sektornr-dblsb[cvfnr].s_datastart)&15);
    shiftval=0; /* unused - avoids warning */ 
  }
  
  lock_bitfat();
  
  min_time=bitfat[0].a_time;
  min_acc=bitfat[0].a_acc;
  merk_i=0;
  /* find area and cvfnr in cache */
  for(i=0;i<BITFATCACHESIZE;++i)
  { if(bitfat[i].a_time<min_time ||
        (bitfat[i].a_time==min_time&&bitfat[i].a_acc<min_acc) )
    { min_time=bitfat[i].a_time;
      min_acc=bitfat[i].a_acc;
      merk_i=i;
    }
    if(bitfat[i].a_buffer!=NULL&&area==bitfat[i].a_area&&
       cvfnr==bitfat[i].a_cvfnr)
    { /* found */
      if(bitfat[i].a_time==CURRENT_TIME)++bitfat[i].a_acc;
      else
      { bitfat[i].a_time=CURRENT_TIME;
        bitfat[i].a_acc=0;
      }
      merk_i=i;
      goto bitfat_okay;      /************* goto ************************/
    }
  }
  /* merk_i = least recently used entry number */
  read_bitfat_sector(sb,area,merk_i,cvfnr);

bitfat_okay:
  if(dblsb[cvfnr].s_cvf_version<STAC3)
  { /* dblspace/drvspace */
    pp=&(bitfat[merk_i].a_buffer->b_data[offset]);
    res=CHS(pp);
    if(new)
    { newval= (*new) ? (res|bitmask) : (res&~bitmask);
      pp[0]=newval;pp[1]=newval>>8;
      bh_dirty(sb,bitfat[merk_i].a_buffer);
    }
    res=(res&bitmask) ? 1 : 0;
  }
  else if(dblsb[cvfnr].s_cvf_version>=STAC3)
  { /* stacker 3 / 4 */
    res=(bitfat[merk_i].a_buffer->b_data[offset]>>shiftval)&bitmask;
    if(new)
    { bitfat[merk_i].a_buffer->b_data[offset]&=~(bitmask<<shiftval);
      bitfat[merk_i].a_buffer->b_data[offset]|=(*new&bitmask)<<shiftval;
      bh_dirty(sb,bitfat[merk_i].a_buffer);
    }
  }
  unlock_bitfat();
  return res;
}


int dbltest(struct inode* inode)
{ int i;

  if(lowest_never_used_cvfnr==0)return 0; /* dmsdos was never in action, 
                                             values are undefined
                                             (this is only for umsdos since
                                             umsdos may call this without
                                             dbl_init being called before) */

  /* am I the inode of a CVF? */
  if(inode->i_ino==0)return 0; /* there is no zero ino for dmsdos */
  for(i=0;i<MAXDBLFILES;++i) 
  { if(dbl_cvf_inos[i]==inode->i_ino&&dbl_sb_dev[i]==inode->i_sb->s_dev)
       return (i+1)<<DMSDOS_CVFNR_BITS;
  }
  return 0;
}
                                           
int dostest(struct inode* inode)
{ /* am I a real DOS inode (=1) or an inode of or in a CVF (=0) */
  if(inode->i_ino<=DMSDOS_ORIGINO_MASK&&dbltest(inode)==0)
  { LOG_FS("DMSDOS: dostest result=1\n");
    return 1;
  }
  LOG_FS("DMSDOS: dostest result=0\n");
  return 0;
}

/* new inode numbering scheme */

/*   --------++++++++--------++++++++
     3322222222221111111111
     10987654321098765432109876543210
     uvvvvvccccccccccccccccsssssspppp
     
     u: unused, always 0 (negative ino means error for a lot of functions)
     v: cvfnr+1, zero means inode of uncompressed host (incl. dmsdos main
        directory inode).
     c: cluster (zero mens this inode is for a file/directory in the main
        directory of a CVF)
     s: virtual sector offset
     p: position in sector
*/

int find_free_cvfnr(void)
{ int i;

  for(i=0;i<MAXDBLFILES;++i)
  { if(dbl_cvf_inos[i]==0)return i;
  }
  return -1;
}

int dbl_init(struct super_block* sb, void* datai, int comp,int cf,int ilong,
              int vfat,int low,int repair, int direct, Listtype*list)
{ int i,j;
  int cvfnr;
  struct buffer_head * bh;
  struct msdos_dir_entry * data;
  char dname[]="        .   ";
  int testvers;
  
  printk("DMSDOS filesystem version %d.%d.%d"
#ifndef DBL_WRITEACCESS
         " read-only"
#else    
         " read-write"
#endif
#ifdef SERIALIZE_RW_ACCESS
         " serialized"
#endif
         "\n",
         (DMSDOS_VERSION&0xff0000)>>16,
         (DMSDOS_VERSION&0x00ff00)>>8,DMSDOS_VERSION&0xff);

  if(lowest_never_used_cvfnr==0)
  { /* first call of DMSDOS filesystem, initialising variables */
  
    /* clear rwlist */
    clear_list();
    
    /* init cluster cache */
    ccache_init();
  
    for(i=0;i<MAXDBLFILES;++i)
    {  dbl_cvf_inos[i]=0;
       dbl_sb_dev[i]=0;
    }
    for(i=0;i<MDFATCACHESIZE;++i)
    {  mdfat[i].a_time=0;
       mdfat[i].a_acc=0;
       mdfat[i].a_buffer=NULL;
    }
    for(i=0;i<DFATCACHESIZE;++i)
    {  dfat[i].a_time=0;
       dfat[i].a_acc=0;
       dfat[i].a_buffer=NULL;
    }
    for(i=0;i<BITFATCACHESIZE;++i)
    {  bitfat[i].a_time=0;
       bitfat[i].a_acc=0;
       bitfat[i].a_buffer=NULL;
    }
    lowest_never_used_cvfnr=MAXDBLFILES;
  }
  
  /* check for maximum cluster size to avoid fs corruption */
  if(direct==0&&MSDOS_SB(sb)->cluster_size>DMSDOS_MAX_DOS_SECTPERCLUST)
  {  printk("DMSDOS: msdos cluster size too large, can't continue\n");
     return -1;
  }
  
  if(direct)
  { testvers=direct;
    if((cvfnr=find_free_cvfnr())<0)
    { printk("DMSDOS: Can't mount, too many CVFs.\n");
      return -1;
    }
    
    dbl_cvf_startcluster[cvfnr]=0; /* direct */
    
    /* find out size */
    dblsb[cvfnr].s_dataend=blk_size[MAJOR(sb->s_dev)][MINOR(sb->s_dev)]*2-1;
    
#ifndef DBL_WRITEACCESS
    comp=READ_ONLY
#endif
    /* stacker currently only uncompressed */
    /* if((testvers==2)&&(comp!=READ_ONLY)) comp=UNCOMPRESSED; */
    
    dblsb[cvfnr].s_comp=comp;
    dblsb[cvfnr].s_cfaktor=cf;
    dblsb[cvfnr].s_support_lfn=ilong+4*vfat;
    dbl_cvf_inos[cvfnr]=1;
    dbl_sb_dev[cvfnr]=sb->s_dev;
    
    printk("DMSDOS: directly mounting CVF %d on device %x %s...\n",
           cvfnr+1,dbl_sb_dev[cvfnr],
           dblsb[cvfnr].s_comp==READ_ONLY?"read-only":"read-write");
    if(testvers==1)
    { /* dblspace correction was relocated. Pavel */
      dblsb[cvfnr].s_dataend-=1;
      setup_mdfat(sb,cvfnr,repair);
    }
    else setup_stacker(sb,cvfnr,repair);
    if(((dblsb[cvfnr].s_support_lfn&1)==0)||low!=0)
       dblsb[cvfnr].s_support_lfn|=8;
    
    /* error test for setup_mdfat/setup_stacker */
    if(dbl_cvf_inos[cvfnr]==0&&dbl_sb_dev[cvfnr]==0)
    { printk("DMSDOS: Error mounting CVF %d directly, aborting.\n",cvfnr+1);
      return -1;
    }
    
    return 0;
  } /* if direct */
    
  for(i=0;i<MSDOS_SB(sb)->dir_entries/MSDOS_DPS;++i)
  { bh=read_real_sector(sb,MSDOS_SB(sb)->dir_start+i);
    if(bh==NULL)
    {  printk("DMSDOS: unable to read msdos root directory\n");
       return -1;
    }
    data=(struct msdos_dir_entry*) bh->b_data;
    
    for(j=0;j<MSDOS_DPS;++j)
    { testvers=0;
      if(strncmp(data[j].name,"DRVSPACE",8)==0)testvers=1; 
      if(strncmp(data[j].name,"DBLSPACE",8)==0)testvers=1;
      if(strncmp(data[j].name,"STACVOL ",8)==0)testvers=2;
      if(testvers)
      { if( ( data[j].name[8]>='0'&&data[j].name[8]<='9'
              &&data[j].name[9]>='0'&&data[j].name[9]<='9'
              &&data[j].name[10]>='0'&&data[j].name[10]<='9'
            ) | (testvers==2&&strncmp(data[j].name+8,"DSK",3)==0)
          ){ /* it is a CVF */
             /* find out whether automounting it is desired or not */
            int k;
            int mountit=0;
            
            switch(list->ltype)
            { case LT_EMPTY:
                mountit=1;
                break;
              case LT_EXCLUDE:
                mountit=1;
                /* fall through */
              case LT_INCLUDE:
                for(k=0;k<list->count;++k)
                { if(strncmp(data[j].name+8,list->names[k],3)==0)
                  { mountit=1-mountit;
                    break;
                  }
                }
            }
            
            if(mountit) 
            { 
             if((cvfnr=find_free_cvfnr())>=0)
             { dbl_cvf_startcluster[cvfnr]=data[j].start;
               dblsb[cvfnr].s_dataend=data[j].size/SECTOR_SIZE-1;

#ifndef DBL_WRITEACCESS
               comp=READ_ONLY;
#endif
               /* stacker currently only uncompressed */
               /* if((testvers==2)&&(comp!=READ_ONLY)) comp=UNCOMPRESSED; */

               dblsb[cvfnr].s_comp=comp;
               
               dblsb[cvfnr].s_cfaktor=cf;
               dblsb[cvfnr].s_support_lfn=ilong+4*vfat;
               dbl_cvf_inos[cvfnr]=(MSDOS_SB(sb)->dir_start+i)*MSDOS_DPS+j;
               dbl_sb_dev[cvfnr]=sb->s_dev;
               strncpy(dname,data[j].name,8);
               strncpy(&(dname[9]),&(data[j].name[8]),3);
               printk("DMSDOS: Automounting CVF %s as CVF %d on device %x ino=%d %s...\n",
                      dname,cvfnr+1,dbl_sb_dev[cvfnr],dbl_cvf_inos[cvfnr],
                      dblsb[cvfnr].s_comp==READ_ONLY?"read-only":"read-write");
               if(testvers==1)
               { /* dblspace correction wos relocated. Pavel */
                 /* the last sector seems to be reserved for some purpose... */
                 dblsb[cvfnr].s_dataend-=1;
                 setup_mdfat(sb,cvfnr,repair);
               }
               else setup_stacker(sb,cvfnr,repair);
               if(((dblsb[cvfnr].s_support_lfn&1)==0)||low!=0)
                   dblsb[cvfnr].s_support_lfn|=8; /*uppercase->lowercase on*/
             }
             else printk("DMSDOS: too many CVFs, ignoring\n");
            }
           } 
      }
    }
    bh_free(sb,bh);
  }
  LOG_REST("DMSDOS: msdos_read_super reports data_start=%d cluster_size=%d\n",
         MSDOS_SB(sb)->data_start,MSDOS_SB(sb)->cluster_size);
  
  return 0;
}

void exit_dbl(struct super_block*sb)
{ int cvfnr,j;

  if(lowest_never_used_cvfnr==0)return;
  
  for(cvfnr=0;cvfnr<MAXDBLFILES;++cvfnr)
  { if(dbl_sb_dev[cvfnr]==sb->s_dev)
    { LOG_REST("DMSDOS: CVF %d ino=%d on device %d unmounted.\n",
             cvfnr+1,dbl_cvf_inos[cvfnr],sb->s_dev);
      
      /* kill rwlist entries */
      clear_list_cvfnr(cvfnr);
      
      /* discard/write cached clusters */
      free_ccache_cvfnr(sb,cvfnr);
      
      /* zero out special infos about unmounted cvf */
      dbl_sb_dev[cvfnr]=0;
      dbl_cvf_inos[cvfnr]=0;
      
      /* kill buffers used by unmounted cvf */
      for(j=0;j<MDFATCACHESIZE;++j)
      { if(mdfat[j].a_cvfnr==cvfnr)
        { if(mdfat[j].a_buffer!=NULL)
          { bh_free(sb,mdfat[j].a_buffer); 
            mdfat[j].a_buffer=NULL;
          }
          mdfat[j].a_time=0;
          mdfat[j].a_acc=0;
        }
      }
      for(j=0;j<DFATCACHESIZE;++j)
      { if(dfat[j].a_cvfnr==cvfnr)
        { if(dfat[j].a_buffer!=NULL)
          { bh_free(sb,dfat[j].a_buffer);
            dfat[j].a_buffer=NULL;
          }
          dfat[j].a_time=0;
          dfat[j].a_acc=0;
        }
      }
      for(j=0;j<BITFATCACHESIZE;++j)
      { if(bitfat[j].a_cvfnr==cvfnr)
        { if(bitfat[j].a_buffer!=NULL)
          { bh_free(sb,bitfat[j].a_buffer);
            bitfat[j].a_buffer=NULL;
          }
          bitfat[j].a_time=0;
          bitfat[j].a_acc=0;
        }
      }
      
    }
  }
}

/* reads stacker BITFAT sumary informations */
__u8 *stac_bitfat_sumary(struct super_block*sb,int cvfnr,
                                  struct buffer_head**pbh)
{ int pos,sector;
  struct buffer_head *bh;
  int bitfat2;
  
  bitfat2=dblsb[cvfnr].s_cvf_version>STAC3;
  pos=dblsb[cvfnr].s_dataend-dblsb[cvfnr].s_datastart; /* number of data sectors */
  pos=(pos+(bitfat2?4:8))>>(bitfat2?2:3);
  pos=(pos+0xF)&~0xF;
  sector=pos/SECTOR_SIZE+dblsb[cvfnr].s_mdfatstart; /* here it's AMAP start !!! */
  *pbh=bh=read_dbl_sector(sb,sector,cvfnr);
  if(bh==NULL) return(NULL);
  return(bh->b_data+pos%SECTOR_SIZE);
};

/* prepared for repair of BITFAT */
int stac_simple_check(struct super_block*sb,int cvfnr, int repair)
{
  unsigned char *sect_array;
  int clust,i,j,val,sect;
  int non_lin_alloc;
  Stac_cwalk cw;
  struct buffer_head*bh;
  __u8* pp;
  int alloc_sects;
  int deleted_clusts;
  int bitfat_dirty;
  static __u8 bitfat_up_to_date_fl[4]={0xAA,0xBB,0xAA,0xAA};
  static __u8 bitfat_changing_fl[4]={0xAA,0xBB,0x55,0x55};
  static __u8 inc_tab[4]={1,4,16,64};

#define set_bits(field,nr,val) field[nr>>2]|=val<<((nr&3)<<1)
#define get_bits(field,nr) ((field[nr>>2]>>((nr&3)<<1))&3)
#define inc_bits(field,nr) field[nr>>2]+=inc_tab[nr&3]

  bitfat_dirty=0;
  if(dblsb[cvfnr].s_comp==READ_ONLY) repair=0;

  /* check bitfat mount id */
  
  { if((pp=stac_bitfat_sumary(sb,cvfnr,&bh))==NULL)
    { printk("DMSDOS: simple_check: read BITFAT sumary error\n");
      return -2;
    };
    if(memcmp(pp,bitfat_up_to_date_fl,sizeof(bitfat_up_to_date_fl)))
    { printk("DMSDOS: simple_check: BITFAT dirty and need dutil repair\n");
      for(i=0;i<16;i++) printk("  %02X",(int)pp[i]);
      printk("\n");
      if(repair)
      { memcpy(pp,bitfat_changing_fl,sizeof(bitfat_changing_fl));
        bitfat_dirty=1;
        bh_dirty(sb,bh);
      };
    };
    printk("DMSDOS: Sumary: info1 = %d\n",(int)CHL((pp+4)));
    printk("DMSDOS: Sumary: info2 = %d\n",(int)(CHL((pp+8)))-(0xF<<28));
    bh_free(sb,bh);
  };

  /* check mdfat */
  
  val=dblsb[cvfnr].s_dataend/4 + 1;
  sect_array=(unsigned char*)MALLOC(val);
  if(sect_array==NULL)
  { printk("DMSDOS: simple_check: MDFAT+BITFAT test skipped (no memory)\n");
    return 2;
  }
  for(i=0;i<val;++i)sect_array[i]=0;
  
  deleted_clusts=0;non_lin_alloc=0;
  for(clust=2;clust<=dblsb[cvfnr].s_max_cluster2;++clust)
  { i=stac_cwalk_init(&cw,sb,clust,cvfnr,0);
    if(i>0)
    { if (cw.flags&0x40) deleted_clusts++;
      while((sect=stac_cwalk_sector(&cw))>0)
      { 
        if(sect>dblsb[cvfnr].s_dataend||sect<dblsb[cvfnr].s_datastart)
        { printk("DMSDOS: MDFAT entry invalid in CVF %d (cluster %d, sect %d)\n",
                 cvfnr+1,clust,sect);
        mde_sect_error:
          stac_cwalk_done(&cw);
          FREE(sect_array);
          return -2;
        }
        val=get_bits(sect_array,sect);
        if(val)
        { if(dblsb[cvfnr].s_cvf_version==STAC3||
              (((cw.flags&0xA0)!=0xA0)&&(cw.flen||cw.fcnt)))
          { printk("DMSDOS: MDFAT crosslink in CVF %d detected (cluster %d )\n",
                   cvfnr+1,clust);
            goto mde_sect_error;
          };
          if(((cw.flags&0xA0)!=0xA0)&&!non_lin_alloc)
          { non_lin_alloc++;
            printk("DMSDOS: Interesting MDFAT non-lin subalocation in CVF %d (cluster %d )\n",
                   cvfnr+1,clust);
          };
        };
        inc_bits(sect_array,sect);
      };
      stac_cwalk_done(&cw);
    } 
    else if (i<0)
    { printk("DMSDOS: MDFAT bad allocation  in CVF %d (cluster %d )\n",
              cvfnr+1,clust);
      FREE(sect_array);
      return -2;
    };
  };
  
  /* check bitfat */

  j=0; /* count BITFAT mismatches */
  alloc_sects=0;

  for(i=dblsb[cvfnr].s_datastart;i<=dblsb[cvfnr].s_dataend;++i)
  { if((val=get_bits(sect_array,i))!=0) alloc_sects++;
    if(dbl_bitfat_value(sb,i,cvfnr,NULL)!=val)
    { if(repair)
      { if(!j)
        { /* repair of first mitchmatch in BITFAT */ 
          printk("DMSDOS: Updating BITFAT CVF %d.\n", cvfnr+1);
          if((pp=stac_bitfat_sumary(sb,cvfnr,&bh))==NULL)
          { printk("DMSDOS: simple_check: read BITFAT sumary error\n");
            FREE(sect_array);
            return -3;
          };
          /* mark bitfat as dirty */
          memcpy(pp,bitfat_changing_fl,sizeof(bitfat_changing_fl));
          bitfat_dirty=1;
          bh_dirty(sb,bh);
          bh_free(sb,bh);
        };
      	dbl_bitfat_value(sb,i,cvfnr,&val);
      	j++;
      }
      else
      { printk("DMSDOS: BITFAT mismatches MDFAT in CVF %d (sector %d is %d and should be %d )\n",
             cvfnr+1,i,dbl_bitfat_value(sb,i,cvfnr,NULL),(unsigned)get_bits(sect_array,i));
        j++;
        if(j==MAX_BITFAT_ERRORS)
        { FREE(sect_array);
          printk("DMSDOS: Too many BITFAT mismatches in CVF %d, check aborted.\n",
                  cvfnr+1);
          return -3;
        }
      }
    }
  }
  if(bitfat_dirty)
  { printk("DMSDOS: Updating BITFAT finished %d.\n", cvfnr+1);
    if((pp=stac_bitfat_sumary(sb,cvfnr,&bh))==NULL)
    { printk("DMSDOS: simple_check: read BITFAT sumary error\n");
      FREE(sect_array);
      return -3;
    };
    /* on success mark BITFAT as up to date */
    memcpy(pp,bitfat_up_to_date_fl,sizeof(bitfat_changing_fl));
    bitfat_dirty=1;
    bh_dirty(sb,bh);
    bh_free(sb,bh);
  };
  printk("DMSDOS: Sumary: Count allocated  = %d\n",alloc_sects);
  printk("DMSDOS: Sumary: Deleted clusters = %d\n",deleted_clusts);
  
  FREE(sect_array);
  if(j!=0&&repair==0)return -3;
  return 0;      
}

/* simple fs check routines (DON'T mount obviously damaged filesystems rw)
*/
int simple_check(struct super_block*sb,int cvfnr,int repair)
{ /* unsigned char field[512*1024]; grr... panics (stack overflow) */
  unsigned char *field;
  unsigned char bits[8]={1,2,4,8,16,32,64,128};
  int i,j,val;
  Mdfat_entry mde;

  /* can't repair ro filesystems */
  if(dblsb[cvfnr].s_comp==READ_ONLY)repair=0;
  
#define setbit(nr) field[nr/8]|=bits[nr%8]
#define getbit(nr) (field[nr/8]&bits[nr%8])

  /* check fat */
  
  /* get memory for field */
  val=dblsb[cvfnr].s_max_cluster2/8 + 1;
  field=(unsigned char*)MALLOC(val);
  if(field==NULL)
  { printk("DMSDOS: simple_check aborted (no memory)\n");
    return 1;
  }
  for(i=0;i<val;++i)field[i]=0;
  
  for(i=2;i<=dblsb[cvfnr].s_max_cluster2;++i)
  { val=dbl_fat_nextcluster(sb,i,cvfnr,NULL);
    if(val!=0&&val!=-1)
    {
      if(getbit(val))
      { printk("DMSDOS: FAT crosslink in CVF %d detected (cluster %d)\n",
               cvfnr+1,i);
        FREE(field);
        return -1;
      }
      setbit(val);
    }
  }
  
  FREE(field);

  if(dblsb[cvfnr].s_cvf_version>=STAC3) 
    return stac_simple_check(sb,cvfnr,repair);

  /* check mdfat */
  
  val=dblsb[cvfnr].s_dataend/8 + 1;
  field=(unsigned char*)MALLOC(val);
  if(field==NULL)
  { printk("DMSDOS: simple_check: MDFAT+BITFAT test skipped (no memory)\n");
    return 2;
  }
  for(i=0;i<val;++i)field[i]=0;
  
  for(i=2;i<=dblsb[cvfnr].s_max_cluster2;++i)
  { dbl_mdfat_value(sb,i,cvfnr,NULL,&mde);
    if(mde.flags&2)
    {
      val=mde.sector_minus_1;
      if(val+mde.size_lo_minus_1>=dblsb[cvfnr].s_dataend||
         val+1<dblsb[cvfnr].s_datastart)
      { printk("DMSDOS: MDFAT entry invalid in CVF %d (cluster %d)\n",
               cvfnr+1,i);
        FREE(field);
        return -2;
      }
      for(j=0;j<=mde.size_lo_minus_1;++j)
      { ++val;
        if(getbit(val))
        { printk("DMSDOS: MDFAT crosslink in CVF %d detected\n",cvfnr+1);
          FREE(field);
          return -2;
        }
        setbit(val);
      }
    }
  }
  
  /* check bitfat */

  /* dataend-1 problem corrected above - dmsdos doesn't touch the
     last sector now because it seems to be reserved... */
  
  j=0; /* count BITFAT mismatches */

  for(i=dblsb[cvfnr].s_datastart;i<=dblsb[cvfnr].s_dataend;++i)
  { if(dbl_bitfat_value(sb,i,cvfnr,NULL)!=(getbit(i)?1:0))
    { if(repair)
      { int newval;
      
        if(j==0)printk("DMSDOS: BITFAT mismatches MDFAT in CVF %d, repairing...\n",
                       cvfnr+1);
        newval=(getbit(i)?1:0);
        dbl_bitfat_value(sb,i,cvfnr,&newval);
        ++j;
      }
      else
      {
        printk("DMSDOS: BITFAT mismatches MDFAT in CVF %d (sector %d)\n",
               cvfnr+1,i);
        ++j;
        if(j==MAX_BITFAT_ERRORS)
        { FREE(field);
          printk("DMSDOS: Too many BITFAT mismatches in CVF %d, check aborted.\n",
                  cvfnr+1);
          return -3;
        }
      }
      
    }
  }
  
  FREE(field);
  if(j!=0&&repair==0)return -3;
  return 0;      
}

/* cluster caching ... */
Cluster_head ccache[CCACHESIZE];

struct wait_queue * ccachewait=NULL;
int ccachelock=0;
void lock_ccache(void)
{ while(ccachelock)sleep_on(&ccachewait);
  ccachelock=1;
}
void unlock_ccache(void)
{ ccachelock=0;
  wake_up(&ccachewait);
}

struct wait_queue * fullwait=NULL;

int help_ccache=0;

/* must be called locked */
Cluster_head* find_in_ccache(int clusternr, int cvfnr,Cluster_head**lastfree,
                             Cluster_head**oldest)
{ int i;
  int lastfree_n=-1;
  int oldest_n=-1;
  unsigned int oldest_time=CURRENT_TIME;
  
  for(i=0;i<CCACHESIZE;++i)
  { if(ccache[i].c_flags==C_FREE)lastfree_n=i;
    else 
    { if(ccache[i].c_time<oldest_time&&ccache[i].c_count==0)
      { oldest_n=i;
        oldest_time=ccache[i].c_time;
      }
      if(ccache[i].c_clusternr==clusternr&&ccache[i].c_cvfnr==cvfnr)
      { /* found */
        return &(ccache[i]);
      }
    }
  }
  if(lastfree_n>=0)*lastfree=&(ccache[lastfree_n]);
  /* uhhh, deadlock possible here... we should always have an oldest when
     there's nothing free any more... */
  /* use help_ccache to avoid always choosing the same cluster */
  if(lastfree_n<0&&oldest_n<0)
  { /* search more carefully */
    /* candidates are all those with count==0 */
    for(i=0;i<CCACHESIZE;++i)
    { ++help_ccache; if(help_ccache>=CCACHESIZE)help_ccache=0;
      if(ccache[help_ccache].c_count==0)
      { oldest_n=help_ccache;
        break;
      }
    }
  }
  if(oldest_n>=0)*oldest=&(ccache[oldest_n]);
  return NULL;
}

Cluster_head* u_read_ch(struct super_block*sb, int clusternr, int cvfnr,
                        int doread)
{ Cluster_head*free=NULL;
  Cluster_head*oldest=NULL;
  Cluster_head*actual;
  int fail_trials=0;

retry:
  lock_ccache();
  
  actual=find_in_ccache(clusternr,cvfnr,&free,&oldest);
  if(actual==NULL)
  { if(free==NULL)
    { /* we must discard one */
      if(oldest==NULL)
      { /* oh, all are in use - we must wait */
        unlock_ccache();
        LOG_CCACHE("DMSDOS: cluster cache full, waiting...\n");
        sleep_on(&fullwait);
        goto retry;
      }
      /* we can use oldest */
      /* write it before if it is dirty... */
      if(oldest->c_flags==C_DIRTY_NORMAL)
      { if(dmsdos_write_cluster(sb,oldest->c_data,oldest->c_length,
                             oldest->c_clusternr,0,oldest->c_cvfnr,0)<0)
        { write_failed:
          oldest->c_time=CURRENT_TIME;
          ++fail_trials;
          printk(KERN_EMERG "DMSDOS: cannot write dirty cluster %d cvfnr %d, trying again later. This is %d. warning.\n",
                 oldest->c_clusternr, oldest->c_cvfnr+1, fail_trials);
          if(fail_trials<CCACHESIZE/5){unlock_ccache();goto retry;}
          panic("DMSDOS: Too many write failures, giving up. Driver STOPPED\n");
        }
      }
      if(oldest->c_flags==C_DIRTY_UNCOMPR)
      { if(dmsdos_write_cluster(sb,oldest->c_data,oldest->c_length,
             oldest->c_clusternr,0,oldest->c_cvfnr,1)<0)goto write_failed;
      }
      FREE(oldest->c_data);
      oldest->c_flags=C_FREE;
      free=oldest;
      /* now we have a free one */
    }
    /* we have a free entry */
    free->c_data=MALLOC(dblsb[cvfnr].s_sectperclust*SECTOR_SIZE);
    if(free->c_data==NULL)
    { /* now we have a serious problem: no memory available */
      /* we cannot simply retry since this locks up the system 
         so we hope waiting for the next buffer to free up solves this
         but if there's no buffer any more to free up we hang forever...
         no better idea here... */
      printk("DMSDOS: ccache: no memory, waiting...\n");
      unlock_ccache();
      sleep_on(&fullwait);
      goto retry;
    }
    free->c_count=0;
    free->c_flags=C_USED; /* it's no longer free */
    free->c_clusternr=clusternr;
    free->c_cvfnr=cvfnr;
    free->c_length=0;
    /* now we need the data from disk */
    if(doread)
    { free->c_length=dmsdos_read_cluster(sb,free->c_data,clusternr,cvfnr);
      if(free->c_length<0)
      { /* uh oh, read failed */
        printk("DMSDOS: ccache: read_cluster failed.\n");
        /* throw away */
        FREE(free->c_data);
        free->c_flags=C_FREE;
        unlock_ccache();
        return NULL;
      }
    }
    /* now we have the actual data */
    actual=free;
  }
  
  /* setup time stamp and counter */
  actual->c_time=CURRENT_TIME;
  ++(actual->c_count);
  
  unlock_ccache();
  
  return actual;
}

/* this routine *must* return a cluster with its slack zerod out if the
   cluster doesn't have maximum length */
Cluster_head* ch_read(struct super_block*sb, int clusternr, int cvfnr)
{ Cluster_head*ch;

  LOG_CCACHE("DMSDOS: ch_read cluster %d cvfnr %d\n",clusternr,cvfnr+1);
  ch=u_read_ch(sb,clusternr,cvfnr,1);
  if(ch!=NULL){LOG_CCACHE("DMSDOS: ch_read: count=%d\n",ch->c_count);}
  LOG_CCACHE("DMSDOS: ch_read finished.\n");
  return ch;
}

/* this routine *needn't* zero out the data */
Cluster_head* ch_noread(struct super_block*sb, int clusternr, int cvfnr)
{ Cluster_head*ch;

  LOG_CCACHE("DMSDOS: ch_noread cluster %d cvfnr %d\n",clusternr,cvfnr+1);
  ch=u_read_ch(sb,clusternr,cvfnr,0);
  if(ch!=NULL){LOG_CCACHE("DMSDOS: ch_noread: count=%d\n",ch->c_count);}
  LOG_CCACHE("DMSDOS: ch_noread finished.\n");
  
  return ch;
}

int ch_dirty(struct super_block*sb, Cluster_head*ch, int near, int ucflag)
{    /* Hmm...
        We must ensure that we can safely delay writing of the cluster
        before. This is done by calling the write_cluster function with
        the SIMULATE flag. If there's a problem then it is expected to
        return an error code (return value<0). In that case we fall back to
        the old write-through caching mechanism.
     */

#ifdef WRITE_BACK
     int i;
#endif

     LOG_CCACHE("DMSDOS: ch_dirty cluster %d cvfnr %d flags %d ucflag %d\n",
                ch->c_clusternr,ch->c_cvfnr+1,ch->c_flags,ucflag);
     
     /* do not write free or deleted clusters */
     if(ch->c_flags==C_FREE||ch->c_flags==C_DELETED)return 0;
     
#ifdef WRITE_BACK
     /* call write_cluster with SIMULATE flag (ucflag=2) first */
     i=dmsdos_write_cluster(sb,ch->c_data,ch->c_length,ch->c_clusternr,
                                 near,ch->c_cvfnr,2);
     if(i>=0)
     { /* success - we can safely delay writing this cluster */
       ch->c_flags=ucflag>0?C_DIRTY_UNCOMPR:C_DIRTY_NORMAL;
       return i;
     }
     /* otherwise fall back to the default write-through code */
#endif
     
     return dmsdos_write_cluster(sb,ch->c_data,ch->c_length,ch->c_clusternr,
                                      near,ch->c_cvfnr,ucflag);
}

/* This routine is called if extremely important data (directory structure)
   really must be written when normal ch_dirty already fails. However, there's
   no guarantee that the data can really be written some time later. This
   function is meant as a last resort. Code will panic later if too much
   clusters are in the "dirty error wait queue". */
void ch_dirty_retry_until_success(struct super_block*sb, Cluster_head*ch, 
                                  int near, int ucflag)
{ if(ch->c_flags!=C_FREE&&ch->c_flags!=C_DELETED)
       ch->c_flags=ucflag>0?C_DIRTY_UNCOMPR:C_DIRTY_NORMAL;
}

void ch_free(struct super_block*sb, Cluster_head*ch)
{ 
  --(ch->c_count);
  LOG_CCACHE("DMSDOS: ch_free cluster %d cvfnr %d count_after_free %d\n",
             ch->c_clusternr,ch->c_cvfnr+1,ch->c_count);
  if(ch->c_count==0)
  { /* throw away deleted clusters immediately */
    if(ch->c_flags==C_DELETED)
    { FREE(ch->c_data);
      ch->c_flags=C_FREE;
    } 
    wake_up(&fullwait);
  }
}

void ccache_init()
{ int i;

  for(i=0;i<CCACHESIZE;++i)
  { ccache[i].c_flags=C_FREE;
    ccache[i].c_time=0;
    ccache[i].c_count=0;
  }
}

/* for unmount */
void free_ccache_cvfnr(struct super_block*sb,int cvfnr)
{ int i;

  lock_ccache();
  for(i=0;i<CCACHESIZE;++i)
  { if(ccache[i].c_flags!=C_FREE&&ccache[i].c_cvfnr==cvfnr)
    { 
      /* write it before if it is dirty... */
      if(ccache[i].c_flags==C_DIRTY_NORMAL)
      { /* we do not check for failure since we cannot do anything for it
           at unmount time (except panic, but this would be worse) */
        dmsdos_write_cluster(sb,ccache[i].c_data,ccache[i].c_length,
                             ccache[i].c_clusternr,0,ccache[i].c_cvfnr,UMOUNT_UCFLAG);
      }
      if(ccache[i].c_flags==C_DIRTY_UNCOMPR)
      { dmsdos_write_cluster(sb,ccache[i].c_data,ccache[i].c_length,
                             ccache[i].c_clusternr,0,ccache[i].c_cvfnr,1);
      }

      if(ccache[i].c_count)printk("DMSDOS: ccache: oh oh, freeing busy cluster...\n");
      FREE(ccache[i].c_data);
      ccache[i].c_flags=C_FREE;
      wake_up(&fullwait);
    }
  }
  unlock_ccache();
}

/* for memory management */
void free_idle_ccache(struct super_block*sb)
{ int i;

  lock_ccache();
  for(i=0;i<CCACHESIZE;++i)
  { if(ccache[i].c_flags!=C_FREE&&ccache[i].c_count==0)
    { /* hmm... would be a good idea to make it depend on free system
         memory whether to discard which cached clusters... and how many...
         any ideas? */
      if(CURRENT_TIME-ccache[i].c_time>MAX_CCACHE_TIME)
      { /* throw away */
        /* but write it before if it is dirty... */
        if(ccache[i].c_flags==C_DIRTY_NORMAL)
        { if(dmsdos_write_cluster(sb,ccache[i].c_data,ccache[i].c_length,
               ccache[i].c_clusternr,0,ccache[i].c_cvfnr,0)<0)continue;
        }
        if(ccache[i].c_flags==C_DIRTY_UNCOMPR)
        { if(dmsdos_write_cluster(sb,ccache[i].c_data,ccache[i].c_length,
               ccache[i].c_clusternr,0,ccache[i].c_cvfnr,1)<0)continue;
        }
        FREE(ccache[i].c_data);
        ccache[i].c_flags=C_FREE;
        wake_up(&fullwait);
      }
    }
  }
  unlock_ccache();

  /* wake up forgotten fullwaits - for safety */
  wake_up(&fullwait);
}

/* for cluster cache syncing - writes all currently unused dirty clusters */
void sync_cluster_cache(struct super_block*sb)
{ int i;

  lock_ccache();
  
  for(i=0;i<CCACHESIZE;++i)
  { if(ccache[i].c_flags==C_DIRTY_NORMAL&&ccache[i].c_count==0)
    { if(dmsdos_write_cluster(sb,ccache[i].c_data,ccache[i].c_length,
                           ccache[i].c_clusternr,0,ccache[i].c_cvfnr,0)>=0)
        ccache[i].c_flags=C_USED;
    }
    if(ccache[i].c_flags==C_DIRTY_UNCOMPR&&ccache[i].c_count==0)
    { if(dmsdos_write_cluster(sb,ccache[i].c_data,ccache[i].c_length,
                           ccache[i].c_clusternr,0,ccache[i].c_cvfnr,1)>=0)
        ccache[i].c_flags=C_USED;
    }
  }
      
  unlock_ccache();
}

/* for cluster cache flushing - writes this cluster immediately */
void sync_one_cluster(struct super_block*sb, int clusternr, int cvfnr)
{ int i;

  lock_ccache();
  
  for(i=0;i<CCACHESIZE;++i)
  { /* find it */
    if(ccache[i].c_clusternr==clusternr&&ccache[i].c_cvfnr==cvfnr
       &&ccache[i].c_flags!=C_FREE)
    { /* found */
      /* write back immediately if dirty */
      
      if(ccache[i].c_flags==C_DIRTY_NORMAL)
      { if(dmsdos_write_cluster(sb,ccache[i].c_data,ccache[i].c_length,
                           ccache[i].c_clusternr,0,ccache[i].c_cvfnr,0)>=0)
          ccache[i].c_flags=C_USED;
      }
      if(ccache[i].c_flags==C_DIRTY_UNCOMPR)
      { if(dmsdos_write_cluster(sb,ccache[i].c_data,ccache[i].c_length,
                           ccache[i].c_clusternr,0,ccache[i].c_cvfnr,1)>=0)
          ccache[i].c_flags=C_USED;
      }
    }
  }
      
  unlock_ccache();
}

/* cluster delete - deletes this cluster from the cache and cancels ch_dirty */
void delete_cache_cluster(struct super_block*sb, int clusternr, int cvfnr)
{ int i;

  lock_ccache();
  
  /* inform the cluster handling routines so they can delete pointers in fs */
  dmsdos_write_cluster(sb,NULL,0,clusternr,0,cvfnr,0);
  
  for(i=0;i<CCACHESIZE;++i)
  { /* find it */
    if(ccache[i].c_clusternr==clusternr&&ccache[i].c_cvfnr==cvfnr
       &&ccache[i].c_flags!=C_FREE)
    { /* found */
      /* is it currently in use ? */
      if(ccache[i].c_count)
      { ccache[i].c_clusternr=-1;
        ccache[i].c_cvfnr=-1;
        ccache[i].c_flags=C_DELETED;
      }
      else /* we can throw it away */
      { FREE(ccache[i].c_data);
        ccache[i].c_flags=C_FREE;
      }
    }
  }
      
  unlock_ccache();
}

void log_ccache_statistics()
{ int j;
  int free;
  int used;
  int dirty_n;
  int dirty_u;
  int del;
  
  lock_ccache();
  
  printk("DMSDOS: ccache statistics:\n");
  
  free=0;
  used=0;
  dirty_n=0;
  dirty_u=0;
  del=0;
  
  for(j=0;j<CCACHESIZE;++j)
  { switch(ccache[j].c_flags)
    { case C_FREE: ++free; break;
      case C_USED: ++used; break;
      case C_DIRTY_NORMAL: ++dirty_n; break;
      case C_DIRTY_UNCOMPR: ++dirty_u; break;
      case C_DELETED: ++del; break;
      default: printk("DMSDOS: log_ccache_statistics: cannot happen.\n");
    }
  }
  
  printk("sum: free=%d used=%d dirty_n=%d dirty_u=%d del=%d\n",
         free,used,dirty_n,dirty_u,del);
         
  unlock_ccache();
}
