#include<stdio.h>
#include<string.h>

/* mkdmsdosfs -make CVFNAME LERNFILE SIZE 
   mkdmsdosfs -lern CVFNAME LERNFILE 
*/   

#define DEFAULT_LERNFILE_PATH "/usr/local/lib/dmsdos/"
#define SAVELEN 64

int isshort(char*test)
{ while(*test)
  { if(*test=='/')return 0;
    ++test;
  }
  return 1;
}

void usage(void)
{ printf("Usage:\n");
  printf("mkdmsdosfs -make CVF_NAME LEARNFILE_NAME CVF_SIZE\n");
  printf("mkdmsdosfs -learn CVF_NAME LEARNFILE_NAME\n");
  printf(" -> CVF_SIZE is in units of 512 bytes.\n");
}

int scan(char*string)
{ int zahl=0;

  if(strncmp(string,"0x",2)==0)sscanf(string+2,"%x",&zahl);
  else sscanf(string,"%d",&zahl);
  return zahl;
}

void make(char*cvfname,char*lernfile,int size)
{ FILE*f1;
  FILE*f2;
  int i,sector,scount,datastart,fatstart,last;
  unsigned char superblock[512];
  unsigned char bootblock[512];

  if(size<=0)
  { printf("Size %d is illegal\n",size);
    return;
  }

  f1=fopen(lernfile,"rb");
  if(f1==NULL)
  { printf("File %s not found.\n",lernfile);
    return;
  }
  f2=fopen(cvfname,"wb");
  if(f2==NULL)
  { printf("Can't write file %s\n",cvfname);
    fclose(f1);
    return;
  }
  
  /* zero out superblock and bootblock */
  for(i=0;i<512;++i)
  { superblock[i]=0;
    bootblock[i]=0;
  }
  
  /* read superblock and bootblock from lernfile */
  for(i=0;i<SAVELEN;++i)superblock[i]=fgetc(f1);
  for(i=0;i<SAVELEN;++i)bootblock[i]=fgetc(f1);
  
  /* get start of bootblock */
  sector=superblock[39]+(superblock[40]<<8);

  /* write superblock */
  for(i=0;i<512;++i)fputc(superblock[i],f2);
  
  /* zero out the hole */
  for(scount=1;scount<sector;++scount)
  { for(i=0;i<512;++i)fputc(0,f2);
  }
  
  /* write bootblock */
  for(i=0;i<512;++i)fputc(bootblock[i],f2);

  /* write signature after bootblock */
  fputc(0xf8,f2);
  fputc('D',f2);
  fputc('R',f2);
  for(i=3;i<512;++i)fputc(0,f2);
    
  datastart=sector+superblock[43]+(superblock[44]<<8)+2;
  
  /* zero out the hole */
  for(scount=sector+2;scount<datastart;++scount)
  { for(i=0;i<512;++i)fputc(0,f2);
  }
  
  /* write fat media byte */
  fatstart=sector+superblock[14]+(superblock[15]<<8);
  fseek(f2,fatstart*512,SEEK_SET);
  fputc(0xf8,f2);
  fputc(0xff,f2);
  fputc(0xff,f2);
  if(bootblock[0x3a]=='6')fputc(0xff,f2); /* 16bit fat */
  
  /* determine last sector */
  if(size<scount+1000)size=scount+1000;
  last=size-1;
  
  /* and write it */
  fseek(f2,last*512,SEEK_SET);
  fputc('M',f2);
  fputc('D',f2);
  fputc('R',f2);
  for(i=3;i<512;++i)fputc(0,f2);
  
  fclose(f1);
  fclose(f2);
}

void lern(char*cvfname,char*lernfile)
{ FILE*f1;
  FILE*f2;
  int i,sector;
  unsigned char superblock[512];
  unsigned char bootblock[512];
  
  f1=fopen(cvfname,"rb");
  if(f1==NULL)
  { printf("File %s not found.\n",cvfname);
    return;
  }
  f2=fopen(lernfile,"wb");
  if(f2==NULL)
  { printf("Can't write file %s\n",lernfile);
    fclose(f1);
    return;
  }
  
  /* read superblock */
  for(i=0;i<512;++i)superblock[i]=fgetc(f1);
  
  /* find bootblock */
  sector=superblock[39]+(superblock[40]<<8);
  
  /* read bootblock */
  fseek(f1,sector*512,SEEK_SET);
  for(i=0;i<512;++i)bootblock[i]=fgetc(f1);
  
  /* save important information to lernfile */
  for(i=0;i<SAVELEN;++i)fputc(superblock[i],f2);
  for(i=0;i<SAVELEN;++i)fputc(bootblock[i],f2);
  
  fclose(f1);
  fclose(f2);
}
   
void main(int argc,char*argv[])
{ char lernfile[1024];

  /* check argument range */
  if(argc<=1||argc>5)
  { usage();
    return;
  }
  
  /* setup real lernfile name */
  if(isshort(argv[3]))
  { strcpy(lernfile,DEFAULT_LERNFILE_PATH);
    strcat(lernfile,argv[3]);
  }
  else strcpy(lernfile,argv[3]);
  
  /* execute what to do */
  if(strcmp(argv[1],"-make")==0&&argc==5)
                     make(argv[2],lernfile,scan(argv[4]));
  else if((strcmp(argv[1],"-learn")==0||strcmp(argv[1],"-lern")==0)&&argc==4)
                     lern(argv[2],lernfile);
  else usage();
}
